import json
import logging
from typing import Any, Callable, Dict, Hashable, List, Optional, Union

from jsonpath_ng import Index, Fields
from jsonpath_ng.ext.filter import Filter
from jsonpath_ng.ext.parser import ExtentedJsonPathParser

from robot.api.deco import keyword


JSONValue = Optional[Union[str, int, float, bool]]
JSONType = Union[Dict[Hashable, JSONValue], List[JSONValue], JSONValue]


class RPAFilter(Filter):
    """Extends default filtering JSON path logic."""

    def filter(self, fn: Callable[[JSONType], bool], data: JSONType) -> JSONType:
        for datum in reversed(self.find(data)):
            index_obj = datum.path
            if isinstance(data, dict):
                index_obj.index = list(data)[index_obj.index]
            index_obj.filter(fn, data)
        return data


class RPAJsonPathParser(ExtentedJsonPathParser):
    """Extends the default JSON path parser found in `jsonpath_ng.ext`."""

    def p_filter(self, p):
        """filter : '?' expressions"""
        p[0] = RPAFilter(p[2])


def parse(path: str, debug: bool = False) -> RPAJsonPathParser:
    return RPAJsonPathParser(debug=debug).parse(path)


class JSON:
    r"""`JSON` is a library for manipulating `JSON`_ files and strings.

    JSON is a common data interchange format inspired by a subset of
    the Javascript programming language, but these days is a de facto
    standard in modern web APIs and is language agnostic.

    .. _JSON: http://json.org/

    Serialization
    =============

    The term `serialization` refers to the process of converting
    Robot Framework or Python types to JSON or the other way around.

    Basic types can be easily converted between the domains,
    and the mapping is as follows:

    ============= =======
    JSON          Python
    ============= =======
    object        dict
    array         list
    string        str
    number (int)  int
    number (real) float
    true          True
    false         False
    null          None
    ============= =======

    About JSONPath
    ==============

    Reading and writing values from/to JSON serializable objects is done
    using `JSONPath`_. It's a syntax designed to quickly and easily refer to
    specific elements in a JSON structure. The specific flavor used in this
    library is based on `jsonpath-ng`_.

    Compared to Python's normal dictionary access, JSONPath expressions can
    target multiple elements through features such as conditionals and wildcards,
    which can simplify many JSON-related operations. It's analogous to XPath
    for XML structures.

    .. _JSONPath: http://goessner.net/articles/JsonPath/
    .. _jsonpath-ng: https://pypi.org/project/jsonpath-ng/#description

    Syntax example
    --------------

    For this example consider the following structure:

    .. code-block:: json

       {
         "clients": [
           {
             "name": "Johnny Example",
             "email": "john@example.com",
             "orders": [
                 {"address": "Streetroad 123", "price": 103.20},
                 {"address": "Streetroad 123", "price": 98.99}
             ]
           },
           {
             "name": "Jane Example",
             "email": "jane@example.com",
             "orders": [
                 {"address": "Waypath 321", "price": 22.00},
                 {"address": "Streetroad 123", "price": 2330.01}
             ]
           }
         ]
       }

    In the simplest case JSONPath can replace nested access:

    .. code-block:: robotframework

        *** Tasks ***
        Nested access
            # First order of first client, with direct dictionary access
            ${value}=    Set variable    ${json}["clients"][0]["orders"][0]

            # JSONPath access
            ${value}=    Get value from JSON    ${json}    $.clients[0].orders[0]

    But the power comes from complicated expressions:

    .. code-block:: robotframework

        *** Tasks ***
        Complicated expressions
            # Find delivery addresses for all orders
            ${prices}=        Get values from JSON    $..address

            # Find orders that cost over 100
            ${expensives}=    Get values from JSON    $..orders[?(@.price>100)]


    Supported Expressions
    ---------------------

    The supported syntax elements are:

    =======================    ===========
    Element                    Description
    =======================    ===========
    ``$``                      Root object/element
    ``@``                      Current object/element inside expressions
    ``.`` or ``[]``            Child operator
    ``..``                     Recursive descendant operator
    ````parent````             Parent operator, see `functions`_
    ``*``                      Wilcard, any element
    ``,``                      Select multiple fields
    ``[n]``                    Array index
    ``[a:b:c]``                Array slice (start, end, step)
    ``[a,b]``                  Union of indices or names
    ``[?()]``                  Apply a filter expression
    ``()``                     Script expression
    ``[\\field]``              Sort descending by ``field``, cannot be combined with
                               filters.
    ``[/field]``               Sort ascending by ``field``, cannot be combined with
                               filters.
    ````str()````              Convert value to string, see `functions`_
    ````sub()````              Regex substitution function, see `functions`_
    ````len````                Calculate value's length, see `functions`_
    ````split()````            String split function, see `functions`_
    ``+`` ``-`` ``*`` ``/``    Arithmetic functions, see `functions`_
    =======================    ===========

    Functions
    ^^^^^^^^^

    This library allows JSON path expressions to include certain functions
    which can provide additional benefit to users. These functions are
    generally encapsulated in backticks (`````). Some functions require
    you to pass arguments similar to a Python function.

    For example, let's say a JSON has nodes on the JSON path
    ``$.books[*].genres`` which are represented as strings of genres with
    commas separating each genre. So for one book, this node might have a
    value like ``horror,young-adult``. You can return a list of first genre
    for each book by using the ``split`` function like so:

    .. code-block:: robotframework

        *** Task ***
        Get genres
            ${genres}=  Get values from JSON    $.books[*].genres.`split(,, 0, -1)`

    Each functions parameters are defined here:

    ===================================  =====
    Function                             Usage
    ===================================  =====
    ``str()``                            No parameters, but parenthesis are required
    ``sub(/regex/, repl)``               The regex pattern must be provided in *regex*
                                         and the replacement value provided in *repl*
    ``len``                              No parameters and no parenthesis
    ``split(char, segment, max_split)``  Separator character provided as *char*, which
                                         index from the resulting array to be returns
                                         provided as *segment*, and maximum number of
                                         splits to perform provided as *max_split*,
                                         ``-1`` for all splits.
    ``parent``                           No parameters, no parenthesis
    ===================================  =====

    **Arithmetic Functions**

    JSON Path can be written and combined to concatenate string values
    or perform arithmetic functions on numerical values. Each JSONPath
    expression used must return the same type, and when performing
    such functions between returned lists, each list must be the same
    length. An example is included in documentation for the keyword
    \`Get values from JSON\`.

    Additional Information
    ^^^^^^^^^^^^^^^^^^^^^^

    There are a multitude of different script expressions
    in addition to the elements listed above, which can
    be seen in the `aforementioned article`__.

    For further library usage examples, see the individual keywords.

    __ JSONPath_
    """

    # TODO: Add more logging about affected rows, at least on debug level

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @keyword("Load JSON from file")
    def load_json_from_file(self, filename: str, encoding="utf-8") -> JSONType:
        """Load JSON data from a file, and return it as JSON serializable object.
        Depending on the input file the object can be either a dictionary,
        a list, or a scalar value.

        :param filename: path to input file
        :param encoding: file character encoding
        :return: JSON serializable object of the JSON file

        Example:

        .. code:: robotframework

            *** Task ***
            Load json
                &{auth}=    Load JSON from file    auth.json
                Log   Current auth token: ${auth.token}

        """
        self.logger.info("Loading JSON from file: %s", filename)
        with open(filename, "r", encoding=encoding) as json_file:
            return json.load(json_file)

    @keyword("Save JSON to file")
    def save_json_to_file(
        self,
        doc: JSONType,
        filename: str,
        indent: Optional[int] = None,
        encoding: str = "utf-8",
    ) -> None:
        """Save a JSON serializable object or a string containing
        a JSON value into a file.

        :param doc: JSON serializable object or string
        :param filename: path to output file
        :param indent: if given this value is used for json file indent
        :param encoding: file character encoding

        Robot Framework Example:

        .. code:: robotframework

            *** Tasks ***
            Save dictionary to file
                ${john}=    Create dictionary    name=John    mail=john@example.com
                Save JSON to file    ${john}    john.json

            Save string to file
                ${mark}=    Set variable    {"name": "Mark", "mail": "mark@example.com"}
                Save JSON to file    ${mark}    mark.json

        Python Example:

        .. code:: python

            from RPA.JSON import JSON

            # Save dictionary to file.
            john = {"name": "John", "mail": "john@example.com"}
            JSON().save_json_to_file(john, "john.json")

        """
        self.logger.info("Saving JSON to file: %s", filename)
        extra_args = {}
        if indent:
            extra_args["indent"] = indent
        doc = self.convert_string_to_json(doc) if isinstance(doc, str) else doc
        with open(filename, "w", encoding=encoding) as outfile:
            json.dump(doc, outfile, **extra_args)

    @keyword("Convert JSON to String")
    def convert_json_to_string(self, doc: JSONType) -> str:
        """Convert a JSON serializable object to a string and return it.

        :param doc: JSON serializable object
        :return: string of the JSON serializable object

        Robot Framework Example:

        .. code:: robotframework

            *** Task ***
            Convert to string
                ${obj}=    Create dictionary    Key=Value
                ${json}=   Convert JSON to string    ${obj}
                Should be equal    ${json}     {"Key": "Value"}

        Python Example:

        .. code:: python

            from RPA.JSON import JSON
            from robot.libraries.BuiltIn import BuiltIn

            obj = {"Key": "Value"}
            json = JSON().convert_json_to_string(obj)
            BuiltIn().should_be_equal(json, '{"Key": "Value"}')

        """
        return json.dumps(doc)

    @keyword("Convert String to JSON")
    def convert_string_to_json(self, doc: str) -> JSONType:
        """Convert a string to a JSON serializable object and return it.

        :param doc: JSON string
        :return: JSON serializable object of the string

        Robot Framework Example:

        .. code:: robotframework

            *** Task ***
            Convert to json
                ${json}=    Set variable    {"Key": "Value"}
                &{obj}=     Convert string to JSON    ${json}
                Should be equal    ${obj.Key}    Value

        Python Example:

        .. code:: python

            from RPA.JSON import JSON
            from robot.libraries.BuiltIn import BuiltIn

            json = '{"Key": "Value"}'
            obj = JSON().convert_string_to_json(json)
            BuiltIn().should_be_equal(obj["Key"], "Value")

        """
        return json.loads(doc)

    @keyword("Add to JSON")
    def add_to_json(self, doc: JSONType, expr: str, value: JSONType) -> JSONType:
        """Add items into a JSON serializable object and return the result.

        If the target is a list, the values are appended to the end.
        If the target is a dict, the keys are either added or updated.

        :param doc: JSON serializable object
        :param expr: JSONPath expression
        :param value: values to either append or update
        :return: JSON serializable object of the updated JSON

        Robot Framework Example:

        .. code:: robotframework

            *** Task ***
            Change the name value for all people
                &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
                &{person}=    Create dictionary      Name=John
                &{after}=     Add to JSON    ${before}   $.People    ${person}

        Python Example:

        .. code:: python

            from RPA.JSON import JSON

            # Change the name value for all people
            js = JSON()
            before = js.convert_string_to_json('{"People": [{"Name": "Mark"}, {"Name": "Jane"}]}')
            person = {"Name": "John"}
            after = js.add_to_json(before, "$.People", person)

            print(after)

        """  # noqa: E501
        self.logger.info("Add to JSON with expression: %r", expr)
        for match in parse(expr).find(doc):
            if isinstance(match.value, dict):
                match.value.update(value)
            if isinstance(match.value, list):
                match.value.append(value)
        return doc

    @keyword("Get value from JSON")
    def get_value_from_json(
        self, doc: JSONType, expr: str, default: Optional[Any] = None
    ) -> str:
        """Get a single value from a JSON serializable object that matches the given expression.

        Raises a ValueError if there is more than one match.
        Returns the given default argument (or None) if there
        were no matches.

        :param doc: JSON serializable object or string
        :param expr: jsonpath expression
        :param default: default value to return in the absence of a match
        :return: string containing the match OR `default` if there are no matches
        :raises ValueError: if more than one match is discovered

        Short Robot Framework Example:

        .. code:: robotframework

            *** Task ***
            Get the name value for the first person
                &{people}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
                ${first}=     Get value from JSON      ${people}   $.People[0].Name

        Short Python Example:

        .. code:: python

            from RPA.JSON import JSON

            # Get the name value for the second person.
            people = {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
            second = JSON().get_value_from_json(people, "$.People[1].Name")
            print(second)

        Extended Robot Framework Example:

        .. code:: robotframework

            *** Settings ***
            Library         RPA.JSON
            Suite Setup     Ingest JSON

            *** Variables ***
            ${JSON_STRING}      {
            ...                   "clients": [
            ...                     {
            ...                       "name": "Johnny Example",
            ...                       "email": "john@example.com",
            ...                       "orders": [
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 103.20, "id":"guid-001"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 98.99, "id":"guid-002"}
            ...                       ]
            ...                     },
            ...                     {
            ...                       "name": "Jane Example",
            ...                       "email": "jane@example.com",
            ...                       "orders": [
            ...                         {"address": "Waypath 321", "state": "WA", "price": 22.00, "id":"guid-003"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 2330.01, "id":"guid-004"},
            ...                         {"address": "Waypath 321", "state": "WA", "price": 152.12, "id":"guid-005"}
            ...                       ]
            ...                     }
            ...                   ]
            ...                 }
            ${ID}               guid-003

            *** Tasks ***
            Get email for specific order id
                ${email}=    Get value from json    ${JSON_DOC}    $.clients[?(@..id=="${ID}")].email
                Log    \\nOUTPUT IS\\n ${email}    console=${True}
                Should be equal as strings    ${email}    jane@example.com

            *** Keywords ***
            Ingest JSON
                ${doc}=    Convert string to json    ${JSON_STRING}
                Set suite variable    ${JSON_DOC}    ${doc}

        """  # noqa: E501
        self.logger.info("Get value from JSON with expression: %r", expr)
        result = [match.value for match in parse(expr).find(doc)]
        if len(result) > 1:
            raise ValueError(
                "Found {count} matches: {values}".format(
                    count=len(result), values=", ".join(str(r) for r in result)
                )
            )

        return result[0] if result else default

    @keyword("Get values from JSON")
    def get_values_from_json(self, doc: JSONType, expr: str) -> list:
        """Get all values from a JSON serializable object that match the given expression.

        :param doc: JSON serializable object or string
        :param expr: JSONPath expression
        :return: list of values that match

        Short Robot Framework Example:

        .. code:: robotframework

            *** Task ***
            Get all the names for all people
                &{people}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
                @{names}=     Get values from JSON     ${people}   $.People[*].Name

        Short Python Example:

        .. code:: python

            from RPA.JSON import JSON

            # Get all the names for all people
            people = {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
            names = JSON().get_values_from_json(people, "$.People[*].Name")
            print(second)

        Extended Robot Framework Example:

        .. code:: robotframework

            *** Settings ***
            Library         RPA.JSON
            Suite Setup     Ingest JSON

            *** Variables ***
            ${JSON_STRING}      {
            ...                   "clients": [
            ...                     {
            ...                       "name": "Johnny Example",
            ...                       "email": "john@example.com",
            ...                       "orders": [
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 103.20, "id":"guid-001"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 98.99, "id":"guid-002"}
            ...                       ]
            ...                     },
            ...                     {
            ...                       "name": "Jane Example",
            ...                       "email": "jane@example.com",
            ...                       "orders": [
            ...                         {"address": "Waypath 321", "state": "WA", "price": 22.00, "id":"guid-003"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 2330.01, "id":"guid-004"},
            ...                         {"address": "Waypath 321", "state": "WA", "price": 152.12, "id":"guid-005"}
            ...                       ]
            ...                     }
            ...                   ]
            ...                 }
            ${ID}               guid-003

            *** Tasks ***
            Get All Prices and Order Ids
                # Arithmetic operations only work when lists are of equal lengths and types.
                ${prices}=    Get values from json
                ...    ${JSON_DOC}
                ...    $.clients[*].orders[*].id + " has price " + $.clients[*].orders[*].price.`str()`
                Log    \\nOUTPUT IS\\n ${prices}    console=${True}
                Should be equal as strings    ${prices}
                ...    ['guid-001 has price 103.2', 'guid-002 has price 98.99', 'guid-003 has price 22.0', 'guid-004 has price 2330.01', 'guid-005 has price 152.12']

            Find Only Valid Emails With Regex
                # The regex used in this example is simplistic and
                # will not work with all email addresses
                ${emails}=    Get values from json
                ...    ${JSON_DOC}
                ...    $.clients[?(@.email =~ "[a-zA-Z]+@[a-zA-Z]+\\.[a-zA-Z]+")].email
                Log    \\nOUTPUT IS\\n ${emails}    console=${True}
                Should be equal as strings    ${emails}    ['john@example.com', 'jane@example.com']

            Find Orders From Texas Over 100
                # The regex used in this example is simplistic and
                # will not work with all email addresses
                ${orders}=    Get values from json
                ...    ${JSON_DOC}
                ...    $.clients[*].orders[?(@.price > 100 & @.state == "TX")]
                Log    \\nOUTPUT IS\\n ${orders}    console=${True}
                Should be equal as strings    ${orders}
                ...    [{'address': 'Streetroad 123', 'state': 'TX', 'price': 103.2, 'id': 'guid-001'}, {'address': 'Streetroad 123', 'state': 'TX', 'price': 2330.01, 'id': 'guid-004'}]


            *** Keywords ***
            Ingest JSON
                ${doc}=    Convert string to json    ${JSON_STRING}
                Set suite variable    ${JSON_DOC}    ${doc}

        """  # noqa: E501
        self.logger.info("Get values from JSON with expression: %r", expr)
        return [match.value for match in parse(expr).find(doc)]

    @keyword("Update value to JSON")
    def update_value_to_json(
        self, doc: JSONType, expr: str, value: JSONType
    ) -> JSONType:
        """Update existing values in a JSON serializable object and return the result.
        Will change all values that match the expression.

        :param doc: JSON or string
        :param expr: JSONPath expression
        :param value: New value for the matching item(s)
        :return: JSON serializable object with updated results

        Short Robot Framework Example:

        .. code:: robotframework

            *** Tasks ***
            Change the name key for all people
                &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
                &{after}=     Update value to JSON     ${before}   $.People[*].Name    JohnMalkovich

        .. code:: python

            from RPA.JSON import JSON

            # Change the name key for all people
            before = {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
            after = JSON().update_value_to_json(before, "$.People[*].Name","JohnMalkovich")
            print(after)

        Extended Robot Framework Example:

        .. code:: robotframework

            *** Settings ***
            Library         RPA.JSON
            Library    Collections
            Suite Setup     Ingest JSON

            *** Variables ***
            ${JSON_STRING}      {
            ...                   "clients": [
            ...                     {
            ...                       "name": "Johnny Example",
            ...                       "email": "john@example.com",
            ...                       "id": "user-001",
            ...                       "orders": [
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 103.20, "id":"guid-001"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 98.99, "id":"guid-002"}
            ...                       ]
            ...                     },
            ...                     {
            ...                       "name": "Jane Example",
            ...                       "email": "jane@example.com",
            ...                       "id": "user-002",
            ...                       "orders": [
            ...                         {"address": "Waypath 321", "state": "WA", "price": 22.00, "id":"guid-003"},
            ...                         {"address": "Streetroad 123", "state": "TX", "price": 2330.01, "id":"guid-004"},
            ...                         {"address": "Waypath 321", "state": "WA", "price": 152.12, "id":"guid-005"}
            ...                       ]
            ...                     }
            ...                   ]
            ...                 }
            ${ID}               guid-003

            *** Tasks ***
            Update user email
                ${updated_doc}=    Update value to json
                ...    ${JSON_DOC}
                ...    $.clients[?(@.id=="user-001")].email
                ...    johnny@example.com
                Log    \\nNEW JSON IS\\n ${updated_doc}    console=${True}
                ${new_email}=    Get value from json    ${updated_doc}    $.clients[?(@.id=="user-001")].email
                Should be equal as strings    ${new_email}    johnny@example.com

            Add additional charge to all prices in WA
                # This example also shows how the update keyword changes the original JSON doc in memory.
                ${id_price}=    Get values from json
                ...    ${JSON_DOC}
                ...    $.clients[*].orders[?(@.state=="WA")].id,price
                FOR    ${order_id}    ${price}    IN    @{id_price}
                    Update value to json    ${JSON_DOC}    $.clients[*].orders[?(@.id=="${order_id}")].price    ${{${price} * 1.06}}
                END
                Log    \\nNEW JSON IS\\n ${JSON_DOC}    console=${True}
                ${one_price}=    Get value from json    ${JSON_DOC}    $..orders[?(@.id==${ID})].price
                Should be equal as numbers    ${one_price}    23.32

            *** Keywords ***
            Ingest JSON
                ${doc}=    Convert string to json    ${JSON_STRING}
                Set suite variable    ${JSON_DOC}    ${doc}

        """  # noqa: E501
        self.logger.info("Update JSON with expression: %r", expr)
        for match in parse(expr).find(doc):
            path = match.path
            if isinstance(path, Index):
                match.context.value[match.path.index] = value
            elif isinstance(path, Fields):
                match.context.value[match.path.fields[0]] = value
        return doc

    @keyword("Delete from JSON")
    def delete_from_json(self, doc: JSONType, expr: str) -> JSONType:
        """Delete values from a JSON serializable object and return the result.
        Will delete all values that match the expression.

        :param doc: JSON serializable object or string
        :param expr: JSONPath expression
        :return: JSON serializable object with values removed

        Example:

        .. code:: robotframework

            *** Task ***
            Delete all people
                &{before}=    Convert string to JSON   {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
                &{after}=     Delete from JSON    ${before}   $.People[*]

        .. code:: python

            from RPA.JSON import JSON

            # Delete all people
            before = {"People": [{"Name": "Mark"}, {"Name": "Jane"}]}
            after = JSON().delete_from_json(before, "$.People[*]")
            print(after)

        """  # noqa: E501
        self.logger.info("Delete from JSON with expression: %r", expr)
        return parse(expr).filter(lambda _: True, doc)



import argparse
import logging
import sys
import traceback
from RPA.Crypto import Crypto, Hash


DESCRIPTION = """
Command-line utility for generating encryption keys, and
hashing, encrypting, and decrypting data. Commonly
used with the RPA.Crypto library.
"""


def create_parser():
    """Create argument parser with all available subcommands."""
    parser = argparse.ArgumentParser(
        description=DESCRIPTION.strip(), formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="be more talkative"
    )

    subparsers = parser.add_subparsers(title="available commands")
    key_parser(subparsers)
    hash_parser(subparsers)
    encrypt_parser(subparsers)
    decrypt_parser(subparsers)

    return parser


def key_args(parser):
    """Add common argument group for providing encryption keys."""
    key = parser.add_argument_group("encryption key arguments")
    mxg = key.add_mutually_exclusive_group(required=True)
    mxg.add_argument("-t", "--text", help="encryption key as text")
    mxg.add_argument("-f", "--file", help="encryption key as file")
    mxg.add_argument("-s", "--secret", help="encryption key as Robocorp Vault secret")


def load_key(args):
    """Parse encryption key arguments into a Crypto library instance."""
    lib = Crypto()

    if args.text:
        lib.use_encryption_key(args.text)
    elif args.file:
        with open(args.file, encoding="utf-8") as infile:
            lib.use_encryption_key(infile.read())
    elif args.secret:
        name, _, key = args.secret.partition(".")
        lib.use_encryption_key_from_vault(name, key)
    else:
        raise RuntimeError("Unhandled encryption key type")

    return lib


def key_parser(parent):
    """Create parser for 'key' subcommand."""
    parser = parent.add_parser("key", help="generate encryption key")
    parser.set_defaults(func=key_command)


def key_command(args):
    """Execute 'key' subcommand."""
    del args  # unused

    key = Crypto().generate_key()
    print(key)

    logging.warning(
        "\nNOTE: Store the generated key in a secure place!"
        "\nIf the key is lost, the encrypted data can not be recovered."
        "\nIf anyone else gains access to it, they can decrypt your data."
    )


def hash_parser(parent):
    """Create parser for 'hash' subcommand."""
    parser = parent.add_parser("hash", help="calculate hash digest")
    parser.set_defaults(func=hash_command)
    parser.add_argument(
        "input", nargs="?", type=argparse.FileType("r"), default=sys.stdin
    )
    parser.add_argument(
        "-m",
        "--method",
        choices=[h.name for h in Hash],
        default="SHA1",
        help="hashing method (default: %(default)s)",
    )


def hash_command(args):
    """Execute 'hash' subcommand."""
    method = Hash[args.method]
    data = args.input.read()
    digest = Crypto().hash_string(data, method)
    print(digest)


def encrypt_parser(parent):
    """Create parser for 'encrypt' subcommand."""
    parser = parent.add_parser("encrypt", help="encrypt data")
    parser.set_defaults(func=encrypt_command)
    parser.add_argument(
        "input",
        help="path to input file, or stdin",
        nargs="?",
        type=argparse.FileType("rb"),
        default=sys.stdin,
    )
    parser.add_argument(
        "output",
        help="path to output file, or stdout",
        nargs="?",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
    )
    key_args(parser)


def encrypt_command(args):
    """Execute 'encrypt' subcommand."""
    lib = load_key(args)
    token = lib.encrypt_string(args.input.read())
    args.output.write(token)


def decrypt_parser(parent):
    """Create parser for 'decrypt' subcommand."""
    parser = parent.add_parser("decrypt", help="decrypt data")
    parser.set_defaults(func=decrypt_command)
    parser.add_argument(
        "input",
        help="path to input file, or stdin",
        nargs="?",
        type=argparse.FileType("rb"),
        default=sys.stdin,
    )
    parser.add_argument(
        "output",
        help="path to output file, or stdout",
        nargs="?",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
    )
    key_args(parser)


def decrypt_command(args):
    """Execute 'decrypt' subcommand."""
    lib = load_key(args)
    token = lib.decrypt_string(args.input.read(), encoding=None)
    args.output.write(token)


def main():
    parser = create_parser()
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.INFO, format="%(message)s"
    )

    try:
        args.func(args)
    except KeyboardInterrupt:
        logging.warning("Aborted by user")
        sys.exit(1)
    except Exception as exc:  # pylint: disable=broad-except
        logging.debug(traceback.format_exc())
        logging.error("Command failed: %s", exc)
        sys.exit(1)


if __name__ == "__main__":
    main()



import urllib.parse
import pytest

from RPA.Browser.Selenium import Selenium, ensure_scheme

from . import RESOURCES_DIR, temp_filename


@pytest.fixture()
def library():
    lib = Selenium()
    yield lib
    lib.close_all_browsers()


@pytest.mark.skip()
class TestBrowserFunctionality:
    def test_print_to_pdf(self, library):
        testfile = RESOURCES_DIR / "browser_docs.html"
        library.open_available_browser(f"file://{testfile}", headless=True)
        with temp_filename() as tmp_file:
            library.print_to_pdf(tmp_file)
            # TODO: get the text without PDF library dependency
            # text = PDF().get_text_from_pdf(tmp_file)

            # assert "Please explicitly use either RPA.Browser.Selenium" in text[1]

    def test_print_to_pdf_different_from_start_page(self, library):
        startpage = RESOURCES_DIR / "alert.html"
        testfile = RESOURCES_DIR / "browser_docs.html"
        library.open_available_browser(f"file://{startpage}", headless=True)
        with temp_filename() as tmp_file:
            library.go_to(f"file://{testfile}")
            library.print_to_pdf(output_path=tmp_file)
            # TODO: get the text without PDF library dependency
            # text = PDF().get_text_from_pdf(tmp_file)

            # assert "Please explicitly use either RPA.Browser.Selenium" in text[1]

    def test_print_to_pdf_exception_on_non_supported_driver(self, library):
        testfile = RESOURCES_DIR / "browser_docs.html"
        library.open_available_browser(
            f"file://{testfile}", browser_selection="firefox", headless=True
        )

        expected = "PDF printing works only with Chrome/Chromium"

        with pytest.raises(NotImplementedError) as err:
            library.print_to_pdf(output_path=None)

        assert str(err.value) == expected


@pytest.mark.parametrize(
    "url,default,scheme",
    [
        ("https://www.google.com", "https", "https"),
        ("http://www.google.com", "https", "http"),
        ("www.google.com", "https", "https"),
        ("about:config", "https", "about"),
        ("www.google.com", None, ""),
    ],
)
def test_ensure_scheme(url, default, scheme):
    result = ensure_scheme(url, default)
    parsed = urllib.parse.urlsplit(result)
    assert parsed.scheme == scheme



import datetime
import email
import logging
import re
import time
from email import policy  # pylint: disable=no-name-in-module
from multiprocessing import AuthenticationError
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import pytz
from exchangelib import (
    DELEGATE,
    IMPERSONATION,
    OAUTH2,
    UTC,
    Account,
    Configuration,
    Credentials,
    FileAttachment,
    Folder,
    HTMLBody,
    Identity,
    ItemAttachment,
    Mailbox,
    Message,
    OAuth2AuthorizationCodeCredentials,
)
from exchangelib.folders import Inbox
from oauthlib.oauth2 import OAuth2Token

from RPA.Email.common import counter_duplicate_path


EMAIL_CRITERIA_KEYS = {
    "subject": "subject",
    "subject_contains": "subject__contains",
    "body": "body",
    "body_contains": "body__contains",
    "sender": "sender",
    "sender_contains": "sender__contains",
    "before": "datetime_received__range",
    "after": "datetime_received__range",
    "between": "datetime_received__range",
    "category": "categories",
    "category_contains": "categories__contains",
    "importance": "importance",
}


def mailbox_to_email_address(mailbox):
    return {
        "name": mailbox.name if hasattr(mailbox, "name") else "",
        "email_address": mailbox.email_address
        if hasattr(mailbox, "email_address")
        else "",
    }


class NoRecipientsError(ValueError):
    """Raised when email to be sent does not have any recipients, cc or bcc addresses."""  # noqa: E501


class OAuth2Creds(OAuth2AuthorizationCodeCredentials):

    """OAuth2 auth code flow credentials wrapper supporting token state on refresh."""

    def __init__(self, *args, on_token_refresh: Callable, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_token_refresh = on_token_refresh

    def on_token_auto_refreshed(self, access_token: OAuth2Token):
        self._on_token_refresh(access_token)


class Exchange:
    """`Exchange` is a library for sending, reading, and deleting emails.
    `Exchange` is interfacing with Exchange Web Services (EWS).

    For more information about server settings, see
    `this Microsoft support article <https://support.microsoft.com/en-us/office/server-settings-you-ll-need-from-your-email-provider-c82de912-adcc-4787-8283-45a1161f3cc3>`_.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Email.Exchange
        Task Setup  Authorize  username=${ACCOUNT}  password=${PASSWORD}

        *** Variables ***
        ${ACCOUNT}              ACCOUNT_NAME
        ${PASSWORD}             ACCOUNT_PASSWORD
        ${RECIPIENT_ADDRESS}    RECIPIENT
        ${IMAGES}               myimage.png
        ${ATTACHMENTS}          C:${/}files${/}mydocument.pdf

        *** Tasks ***
        Task of sending email
            Send Message  recipients=${RECIPIENT_ADDRESS}
            ...           subject=Exchange Message from RPA Robot
            ...           body=<p>Exchange RPA Robot message body<br><img src='myimage.png'/></p>
            ...           save=${TRUE}
            ...           html=${TRUE}
            ...           images=${IMAGES}
            ...           cc=EMAIL_ADDRESS
            ...           bcc=EMAIL_ADDRESS
            ...           attachments=${ATTACHMENTS}

        Task of listing messages
            # Attachments are saved specifically with a keyword Save Attachments
            ${messages}=    List Messages
            FOR    ${msg}    IN    @{messages}
                Log Many    ${msg}
                ${attachments}=    Run Keyword If    "${msg}[subject]"=="about my orders"
                ...    Save Attachments
                ...    ${msg}
                ...    save_dir=${CURDIR}${/}savedir
            END
            # Using save_dir all attachments in listed messages are saved
            ${messages}=    List Messages
            ...    INBOX/Problems/sub1
            ...    criterion=subject:'about my orders'
            ...    save_dir=${CURDIR}${/}savedir2
            FOR    ${msg}    IN    @{messages}
                Log Many    ${msg}
            END

        Task of moving messages
            Move Messages    criterion=subject:'about my orders'
            ...    source=INBOX/Processed Purchase Invoices/sub2
            ...    target=INBOX/Problems/sub1

    **Python**

    .. code-block:: python

        from RPA.Email.Exchange import Exchange

        ex_account = "ACCOUNT_NAME"
        ex_password = "ACCOUNT_PASSWORD"

        mail = Exchange()
        mail.authorize(username=ex_account, password=ex_password)
        mail.send_message(
            recipients="RECIPIENT",
            subject="Message from RPA Python",
            body="RPA Python message body",
        )

    **About criterion parameter**

    Following table shows possible criterion keys that can be used to filter emails.
    There apply to all keywords which have ``criterion`` parameter.

    ================= ================
    Key               Effective search
    ================= ================
    subject           subject to match
    subject_contains  subject to contain
    body              body to match
    body_contains     body to contain
    sender            sender (from) to match
    sender_contains   sender (from) to contain
    before            received time before this time
    after             received time after this time
    between           received time between start and end
    category          categories to match
    category_contains categories to contain
    importance        importance to match
    ================= ================

    Keys `before`, `after` and `between` at the moment support two
    different timeformats either `%d-%m-%Y %H:%M` or `%d-%m-%Y`. These
    keys also support special string `NOW` which can be used especially
    together with keyword ``Wait for message  criterion=after:NOW``.

    When giving time which includes hours and minutes then the whole
    time string needs to be enclosed into single quotes.

    .. code-block:: bash

        before:25-02-2022
        after:NOW
        between:'31-12-2021 23:50 and 01-01-2022 00:10'

    Different criterion keys can be combined.

    .. code-block:: bash

        subject_contains:'new year' between:'31-12-2021 23:50 and 01-01-2022 00:10'

    Please **note** that all values in the criterion that contain spaces need
    to be enclosed within single quotes.

    In the following example the email `subject` is going to matched
    only against `new` not `new year`.

    .. code-block:: bash

        subject_contains:new year

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.credentials = None
        self.config = None
        self.account = None
        self._saved_attachments = []

    def on_token_refresh(self, token: OAuth2Token):
        """Callable you can override in order to save the newly obtained token in a
        safe place.
        """
        self.logger.info(
            "OAuth2 token was refreshed. (new expiry: %d)", token["expires_at"]
        )

    def authorize(
        self,
        username: str,
        password: Optional[str] = None,
        autodiscover: Optional[bool] = True,
        access_type: Optional[str] = "DELEGATE",
        server: Optional[str] = None,
        primary_smtp_address: Optional[str] = None,
        is_oauth: bool = False,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[dict] = None,
    ) -> None:
        """Connect to Exchange account

        :param username: account username
        :param password: account password (can be skipped with OAuth2)
        :param autodiscover: use autodiscover or set it off (on by default)
        :param access_type: default "DELEGATE", other option "IMPERSONATION"
        :param server: required for configuration setting (with `autodiscover` off)
        :param primary_smtp_address: by default set to username, but can be
            set to be different from username
        :param is_oauth: use the OAuth2 authorization code flow (instead of basic auth)
        :param client_id: registered application ID
        :param client_secret: registered application secret (password)
        :param token: contains access and refresh tokens, type, scope, expiry etc.
        """
        kwargs = {}
        kwargs["autodiscover"] = autodiscover
        kwargs["access_type"] = (
            DELEGATE if access_type.upper() == "DELEGATE" else IMPERSONATION
        )
        kwargs["primary_smtp_address"] = (
            primary_smtp_address if primary_smtp_address else username
        )
        if is_oauth:
            self.credentials = OAuth2Creds(
                on_token_refresh=self.on_token_refresh,
                identity=Identity(upn=username),
                client_id=client_id,
                client_secret=client_secret,
                # Contains at least a non-expired access token or non-revoked refresh
                #  one. (otherwise an authorization code should be present inside)
                access_token=OAuth2Token(params=token) if token else None,
            )
        else:
            if password is None:
                raise ValueError("A password should be provided with basic auth")
            self.credentials = Credentials(username, password)
        if server:
            self.config = Configuration(
                server=server,
                credentials=self.credentials,
                # Automatically detects authentication type based on the provided
                #  `credentials` object. (when not using OAuth2)
                auth_type=OAUTH2 if is_oauth else None,
            )
            kwargs["config"] = self.config
        else:
            kwargs["credentials"] = self.credentials

        if self.config and autodiscover:
            self.logger.warning(
                "Autodiscovery is left ON while using custom configuration, you may "
                "need to turn it OFF in order to use the custom setting."
            )
        self.account = Account(**kwargs)

    def list_messages(
        self,
        folder_name: Optional[str] = None,
        criterion: Optional[str] = None,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
        count: Optional[int] = 100,
        save_dir: Optional[str] = None,
    ) -> list:
        """List messages in the account inbox. Order by descending
        received time.

        :param folder_name: name of the email folder, default INBOX
        :param criterion: list messages matching criterion
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :param count: number of messages to list
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        """
        # pylint: disable=no-member
        messages = []
        source_folder = self._get_folder_object(folder_name)
        if criterion:
            filter_dict = self._get_filter_key_value(criterion)
            items = source_folder.filter(**filter_dict)
        else:
            items = source_folder.all()
        for item in items.order_by("-datetime_received")[:count]:
            attachments = []
            if save_dir and len(item.attachments) > 0:
                # NOTE(cmin764): Default `overwrite` param is assumed here.
                attachments = self._save_attachments(item, save_dir)
            messages.append(self._get_email_details(item, attachments))
        return messages

    def list_unread_messages(
        self,
        folder_name: Optional[str] = None,
        criterion: Optional[str] = None,
        contains: Optional[bool] = False,
        count: Optional[int] = 100,
        save_dir: Optional[str] = None,
    ) -> list:
        """List unread messages in the account inbox. Order by descending
        received time.

        :param folder_name: name of the email folder, default INBOX
        :param criterion: list messages matching criterion
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :param count: number of messages to list
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        """
        messages = self.list_messages(folder_name, criterion, contains, count, save_dir)
        return [m for m in messages if not m["is_read"]]

    def _get_inbox_folder(
        self, folder_name: str, parent_folder: Optional[str] = None
    ) -> Inbox:
        if parent_folder is None or parent_folder is self.account.inbox:
            parent = self.account.inbox
        else:
            parent = self.account.inbox / parent_folder
        return parent / folder_name if folder_name else parent

    def send_message(
        self,
        recipients: Optional[Union[List[str], str]] = None,
        subject: Optional[str] = "",
        body: Optional[str] = "",
        attachments: Optional[Union[List[str], str]] = None,
        html: Optional[bool] = False,
        images: Optional[Union[List[str], str]] = None,
        cc: Optional[Union[List[str], str]] = None,
        bcc: Optional[Union[List[str], str]] = None,
        save: Optional[bool] = False,
    ) -> None:
        """Keyword for sending message through connected Exchange account.

        :param recipients: list of email addresses
        :param subject: message subject, defaults to ""
        :param body: message body, defaults to ""
        :param attachments: list of filepaths to attach, defaults to `None`
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline use, defaults to `None`
        :param cc: list of email addresses
        :param bcc: list of email addresses
        :param save: is sent message saved to Sent messages folder or not,
            defaults to False

        Email addresses can be prefixed with ``ex:`` to indicate an Exchange
        account address.

        At least one target needs to exist for `recipients`, `cc` or `bcc`.
        """
        if not self.account:
            raise AuthenticationError("Not authorized to any Exchange account")
        recipients, cc, bcc, attachments, images = self._handle_message_parameters(
            recipients, cc, bcc, attachments, images
        )
        if not recipients and not cc and not bcc:
            raise NoRecipientsError(
                "Atleast one address is required for 'recipients', 'cc' or 'bcc' parameter"  # noqa: E501
            )
        self.logger.info("Sending message to %s", ",".join(recipients))

        m = Message(
            account=self.account,
            subject=subject,
            body=body,
            to_recipients=recipients,
            cc_recipients=cc,
            bcc_recipients=bcc,
        )

        self._add_attachments_to_msg(attachments, m)
        self._add_images_inline_to_msg(images, html, body, m)

        if html:
            m.body = HTMLBody(body)
        else:
            m.body = body

        # TODO. The exchangelib does not seem to provide any straightforward way of
        # verifying if message was sent or not
        if save:
            m.folder = self.account.sent
            m.send_and_save()
        else:
            m.send()

    def _handle_message_parameters(self, recipients, cc, bcc, attachments, images):
        recipients = recipients or []
        cc = cc or []
        bcc = bcc or []
        attachments = attachments or []
        images = images or []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(cc, list):
            cc = cc.split(",")
        if not isinstance(bcc, list):
            bcc = bcc.split(",")
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")
        if not isinstance(images, list):
            images = str(images).split(",")
        recipients, cc, bcc = self._handle_recipients(recipients, cc, bcc)
        return recipients, cc, bcc, attachments, images

    def _handle_recipients(self, recipients, cc, bcc):
        recipients = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p
            for p in recipients
        ]
        cc = [Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in cc]
        bcc = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in bcc
        ]
        return recipients, cc, bcc

    def _add_attachments_to_msg(self, attachments, msg):
        for attachment in attachments:
            attachment = attachment.strip()
            with open(attachment, "rb") as f:
                atname = str(Path(attachment).name)
                fileat = FileAttachment(name=atname, content=f.read())
                msg.attach(fileat)

    def _add_images_inline_to_msg(self, images, html, body, msg):
        for image in images:
            image = image.strip()
            with open(image, "rb") as f:
                imname = str(Path(image).name)
                fileat = FileAttachment(
                    name=imname, content=f.read(), content_id=imname
                )
                msg.attach(fileat)
                if html:
                    body = body.replace(imname, f"cid:{imname}")

    def create_folder(self, folder_name: str, parent_folder: Optional[str] = None):
        """Create email folder.

        :param folder_name: name for the new folder (required)
        :param parent_folder: name for the parent folder, by default INBOX
        """
        parent = self._get_inbox_folder(folder_name, parent_folder)
        self.logger.info("Create folder %r", folder_name)
        new_folder = Folder(parent=parent, name=folder_name)
        new_folder.save()

    def delete_folder(self, folder_name: str, parent_folder: Optional[str] = None):
        """Delete email folder.

        :param folder_name: current folder name (required)
        :param parent_folder: name for the parent folder, by default INBOX
        """
        folder_to_delete = self._get_inbox_folder(folder_name, parent_folder)
        self.logger.info("Delete folder %r", folder_name)
        folder_to_delete.delete()

    def rename_folder(
        self,
        oldname: str,
        newname: str,
        parent_folder: Optional[str] = None,
    ):
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param parent_folder: name for the parent folder, by default INBOX
        :return: True if operation was successful, False if not
        """
        parent = self._get_inbox_folder("", parent_folder)
        self.logger.info("Rename folder %r to %r", oldname, newname)
        items = self._get_inbox_folder(oldname, parent_folder).all()
        old_folder = Folder(parent=parent, name=oldname)
        old_folder.name = newname
        old_folder.save()
        items.move(to_folder=parent / newname)
        self.delete_folder(oldname, parent_folder)

    def empty_folder(
        self,
        folder_name: str,
        parent_folder: Optional[str] = None,
        delete_sub_folders: Optional[bool] = False,
    ):
        """Empty email folder of all items

        :param folder_name: current folder name (required)
        :param parent_folder: name for the parent folder, by default INBOX
        :param delete_sub_folders: delete sub folders or not, by default False
        :return: True if operation was successful, False if not
        """
        if parent_folder is None:
            empty_folder = self._get_folder_object(folder_name)
        else:
            empty_folder = self._get_folder_object(f"{parent_folder} / {folder_name}")
        self.logger.info("Empty folder %r", empty_folder)
        empty_folder.empty(delete_sub_folders=delete_sub_folders)

    def move_messages(
        self,
        criterion: Optional[str] = "",
        source: Optional[str] = None,
        target: Optional[str] = None,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
    ) -> bool:
        """Move message(s) from source folder to target folder

        :param criterion: move messages matching this criterion
        :param source: source folder
        :param target: target folder
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :return: boolean result of operation, True if 1+ items were moved else False

        Criterion examples:

        - subject:my message subject
        - body:something in body
        - sender:sender@domain.com
        """
        source_folder = self._get_folder_object(source)
        target_folder = self._get_folder_object(target)
        if source_folder == target_folder:
            raise KeyError("Source folder is same as target folder")
        filter_dict = self._get_filter_key_value(criterion)
        items = source_folder.filter(**filter_dict)
        if items and items.count() > 0:
            items.move(to_folder=target_folder)
            return True
        else:
            self.logger.warning("No items match criterion '%s'", criterion)
            return False

    def move_message(
        self,
        msg: Optional[dict],
        target: Optional[str],
    ):
        """Move a message into target folder

        :param msg: dictionary of the message
        :param target: path to target folder
        :raises AttributeError: if `msg` is not a dictionary containing
         `id` and `changekey` attributes

        Example:

        .. code-block:: robotframework

            ${messages}=    List Messages
            ...    INBOX
            ...    criterion=subject:about my orders
            FOR    ${msg}    IN    @{messages}
                Run Keyword If    "${msg}[sender][email_address]"=="${priority_account}"
                ...    Move Message    ${msg}    target=INBOX / Problems / priority
            END
        """
        if not all(k in msg for k in ["id", "changekey"]):
            raise AttributeError(
                "Move Message keyword expects message dictionary "
                'containing "id" and "changekey" attributes'
            )
        message_id = [(msg["id"], msg["changekey"])]
        target_folder = self._get_folder_object(target)
        self.account.bulk_move(ids=message_id, to_folder=target_folder)

    def _get_folder_object(self, folder_name):
        if not folder_name:
            return self.account.inbox

        folders = folder_name.split("/")
        if "inbox" in folders[0].lower():
            folders[0] = self.account.inbox
        folder_object = None
        for folder in folders:
            if folder_object:
                folder_object /= folder.strip()
            else:
                folder_object = folder

        return folder_object

    def _get_filter_key_value(self, criterion):
        regex1 = rf"({':|'.join(EMAIL_CRITERIA_KEYS)}:|or|and)'(.*?)'"
        regex2 = rf"({':|'.join(EMAIL_CRITERIA_KEYS)}:|or|and)(\S*)\s*"
        parts = re.findall(regex1, criterion, re.IGNORECASE)
        valid_filters = {}
        for part in parts:
            res = self._parse_email_criteria(part)
            if not res or len(res) != 2:
                continue
            self.logger.debug("First regex pass: %s %s", res[0], res[1])
            if res[0] not in valid_filters.keys():
                valid_filters[res[0]] = res[1]
        parts = re.findall(regex2, criterion, re.IGNORECASE)
        for part in parts:
            res = self._parse_email_criteria(part)
            if not res or len(res) != 2:
                continue
            self.logger.debug("Second regex pass: %s %s", res[0], res[1])
            if res[0] not in valid_filters.keys():
                valid_filters[res[0]] = res[1]
        if criterion and criterion != "" and len(valid_filters) == 0:
            raise KeyError("Invalid criterion '%s'" % criterion)
        self.logger.info(
            "Using filter: %s",
            ",".join(["{}:{}".format(k, v) for k, v in valid_filters.items()]),
        )
        return valid_filters

    def _parse_email_criteria(self, part):
        original_key, value = part
        original_key = original_key.replace(":", "")
        if original_key in ["and", "or", "contains"]:
            # Note. Not implemented yet
            return None
        key = None
        if original_key in EMAIL_CRITERIA_KEYS.keys():
            key = EMAIL_CRITERIA_KEYS[original_key]
        else:
            raise KeyError("Unknown email criteria key '%s'" % original_key)

        if value.startswith("'") and value.endswith("'"):
            value = value[1:-1]

        value = self._handle_specific_keys(original_key, value)
        self.logger.debug("Returning parsed criteria (%s, %s)", key, value)
        return key, value

    def _parse_date_from_string(self, date_string):
        date = None
        if date_string.upper() == "NOW":
            right_now = datetime.datetime.now()
            return right_now.astimezone(pytz.utc)
        try:
            date = datetime.datetime.strptime(date_string, "%d-%m-%Y %H:%M")
            return self._date_for_exchange(date)
        except ValueError:
            pass
        try:
            date = datetime.datetime.strptime(date_string, "%d-%m-%Y")
            return self._date_for_exchange(date)
        except ValueError:
            pass
        return date

    def _date_for_exchange(self, date):
        return datetime.datetime(
            date.year,
            date.month,
            date.day,
            date.hour,
            date.minute,
            tzinfo=UTC,
        )

    def _handle_specific_keys(self, key, value):
        if key.upper() == "BEFORE":
            start = datetime.datetime(1972, 1, 1, tzinfo=UTC)
            end = self._parse_date_from_string(value)
            value = (start, end)
        if key.upper() == "AFTER":
            start = self._parse_date_from_string(value)
            end = datetime.datetime(2050, 1, 1, tzinfo=UTC)
            value = (start, end)
        if key.upper() == "BETWEEN":
            ranges = value.upper().split(" AND ")
            if len(ranges) == 2:
                start = self._parse_date_from_string(ranges[0])
                end = self._parse_date_from_string(ranges[1])
                value = (start, end)
            else:
                return None
        if key.upper() == "IMPORTANCE":
            try:
                value = value.capitalize()
            except ValueError:
                return None
        return value

    def wait_for_message(
        self,
        criterion: Optional[str] = "",
        timeout: Optional[float] = 5.0,
        interval: Optional[float] = 1.0,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
        save_dir: Optional[str] = None,
    ) -> Any:
        """Wait for email matching `criterion` to arrive into INBOX.

        :param criterion: wait for message matching criterion
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0 (minimum)
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
         THIS PARAMETER IS DEPRECATED AS OF rpaframework 12.9.0
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        :return: list of messages
        """
        self.logger.info("Wait for messages")
        end_time = time.time() + float(timeout)
        filter_dict = self._get_filter_key_value(criterion)
        items = None
        # minimum interval is 1.0 seconds
        interval = max(interval, 1.0)
        while time.time() < end_time:
            items = self.account.inbox.filter(**filter_dict)  # pylint: disable=E1101
            if items.count() > 0:
                break
            time.sleep(interval)
        messages = []
        for item in items:
            attachments = []
            if save_dir and len(item.attachments) > 0:
                # NOTE(cmin764): Default `overwrite` param is assumed here.
                attachments = self._save_attachments(item, save_dir)
            messages.append(self._get_email_details(item, attachments))

        if len(messages) == 0:
            self.logger.info("Did not receive any matching items")
        return messages

    def _save_attachments(
        self,
        item,
        save_dir: str,
        attachments_from_emls: bool = False,
        overwrite: bool = False,
    ):
        self._saved_attachments = self._saved_attachments or []
        incoming_items = item.attachments if hasattr(item, "attachments") else item

        for attachment in incoming_items:
            self.logger.debug("Attachment type: %s", type(attachment))
            local_path = Path(save_dir) / attachment.name
            if not overwrite:
                local_path = counter_duplicate_path(local_path)

            if isinstance(attachment, FileAttachment):
                with open(local_path, "wb") as stream_out, attachment.fp as stream_in:
                    buffer = stream_in.read(1024)
                    while buffer:
                        stream_out.write(buffer)
                        buffer = stream_in.read(1024)
            elif isinstance(attachment, ItemAttachment):
                with open(local_path, "wb") as message_out:
                    message_out.write(attachment.item.mime_content)
                if attachments_from_emls:
                    self._save_attachments(
                        attachment.item,
                        save_dir,
                        attachments_from_emls=False,
                        overwrite=overwrite,
                    )
            else:
                self.logger.warning(
                    "Unrecognized attachment type: %s", type(attachment)
                )
                return self._saved_attachments

            self.logger.info("Attachment saved to: %s", local_path)
            self._saved_attachments.append(
                self._new_attachment_dictionary(attachment, local_path)
            )

        return self._saved_attachments

    @staticmethod
    def _new_attachment_dictionary(
        attachment: Union[FileAttachment, ItemAttachment], local_path: Path
    ) -> Dict:
        return {
            "name": attachment.name,
            "content_type": attachment.content_type,
            "size": attachment.size,
            "is_contact_photo": getattr(attachment, "is_contact_photo", False),
            "local_path": str(local_path),
        }

    def _get_email_details(self, item, attachments):
        return {
            "subject": item.subject,
            "sender": mailbox_to_email_address(item.sender),
            "datetime_received": item.datetime_received,
            "folder": str(self._get_folder_object(item.folder)),
            "body": item.body,
            "text_body": item.text_body,
            "received_by": mailbox_to_email_address(item.received_by),
            "cc_recipients": [mailbox_to_email_address(cc) for cc in item.cc_recipients]
            if item.cc_recipients
            else [],
            "bcc_recipients": [
                mailbox_to_email_address(bcc) for bcc in item.bcc_recipients
            ]
            if item.bcc_recipients
            else [],
            "is_read": item.is_read,
            "importance": item.importance,
            "message_id": item.message_id,
            "size": item.size,
            "categories": item.categories,
            "has_attachments": len(item.attachments) > 0,
            "attachments": attachments,
            "attachments_object": item.attachments,
            "id": item.id,
            "changekey": item.changekey,
            "mime_content": item.mime_content,
        }

    def save_attachments(
        self,
        message: Union[dict, str],
        save_dir: Optional[str] = None,
        attachments_from_emls: bool = False,
        overwrite: bool = False,
    ) -> list:
        """Save attachments from message into given directory.

        :param message: dictionary or .eml file path containing message details
        :param save_dir: file path where attachments will be saved
        :param attachments_from_emls: pass `True` if the attachment is an EML file (for
            saving attachments from that EML file instead), `False` otherwise (default)
        :param overwrite: overwrite existing downloaded attachments with the same name
            if set to `True`, `False` otherwise (default)
        :return: list of saved attachments

        Example:

        .. code:: robotframework

            ${messages} =    List Messages
            FOR    ${msg}    IN    @{messages}
                Save Attachments    ${msg}    %{ROBOT_ARTIFACTS}
                ...    attachments_from_emls=${True}
            END

            ${attachments} =    Save Attachments    ${CURDIR}${/}saved.eml
            ...    %{ROBOT_ARTIFACTS}    overwrite=${True}
        """  # noqa: E501
        self._saved_attachments = []
        if isinstance(message, dict):
            return self._save_attachments(
                message["attachments_object"],
                save_dir,
                attachments_from_emls=attachments_from_emls,
                overwrite=overwrite,
            )
        else:
            # extract attachments from .eml file
            absolute_filepath = Path(message).resolve()
            if absolute_filepath.suffix != ".eml":
                raise ValueError("Filename extension needs to be '.eml'")
            return self._save_attachments_from_file(
                message, save_dir, overwrite=overwrite
            )

    def _save_attachments_from_file(
        self, filename: str, save_dir: str, *, overwrite: bool
    ):
        """
        Try to extract the attachments from given .eml file
        """
        # ensure that an output dir exists
        attachments = []
        with open(filename, "r") as f:  # pylint: disable=unspecified-encoding
            msg = email.message_from_file(  # pylint: disable=no-member
                f, policy=policy.default
            )
            for attachment in msg.iter_attachments():
                output_filename = attachment.get_filename()
                # If no attachments are found, skip this file
                if output_filename:
                    local_path = Path(save_dir) / output_filename
                    if not overwrite:
                        local_path = counter_duplicate_path(local_path)
                    with open(local_path, "wb") as of:
                        payload = attachment.get_payload(decode=True)
                        of.write(payload)
                        attachments.append(
                            {
                                "name": output_filename,
                                "content_type": None,
                                "size": len(payload),
                                "is_contact_photo": None,
                                "local_path": str(local_path),
                            }
                        )
            if len(attachments) == 0:
                self.logger.warning("No attachment found for file %s!", f.name)
        return attachments

    def save_message(self, message: dict, filename: str):
        """Save email as .eml file.

        :param message: dictionary containing message details
        :param filename: name of the file to save message into
        """
        absolute_filepath = Path(filename).resolve()
        if absolute_filepath.suffix != ".eml":
            raise ValueError("Filename extension needs to be '.eml'")
        with open(absolute_filepath, "wb") as message_out:
            message_out.write(message["mime_content"])



import logging
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List

from PIL import Image
from PIL import ImageDraw
from PIL import ImageOps

from RPA.core.geometry import Region, to_point, to_region
from RPA.core.notebook import notebook_image

try:
    from RPA.recognition import templates

    HAS_RECOGNITION = True
except ImportError:
    HAS_RECOGNITION = False


def to_image(obj):
    """Convert `obj` to instance of Pillow's Image class."""
    if obj is None or isinstance(obj, Image.Image):
        return obj
    return Image.open(obj)


def clamp(minimum, value, maximum):
    """Clamp value between given minimum and maximum."""
    return max(minimum, min(value, maximum))


def chunks(obj, size, start=0):
    """Convert `obj` container to list of chunks of `size`."""
    return [obj[i : i + size] for i in range(start, len(obj), size)]


@dataclass
class RGB:
    """Container for a single RGB value."""

    red: int
    green: int
    blue: int

    @classmethod
    def from_pixel(cls, value):
        """Create RGB value from pillow getpixel() return value."""
        # RGB(A), ignore possible alpha channel
        if isinstance(value, tuple):
            red, green, blue = value[:3]
        # Grayscale
        else:
            red, green, blue = [value] * 3

        return cls(red, green, blue)

    def luminance(self) -> int:
        """Approximate (perceived) luminance for RGB value."""
        return (self.red * 2 + self.green * 3 + self.blue) // 6


class ImageNotFoundError(Exception):
    """Raised when template matching fails."""


class Images:
    """`Images` is a library for general image manipulation.
    For image-based desktop automation, use the ``RPA.Desktop`` library.

    **Coordinates**

    The coordinates used in the library are pairs of x and y values that
    represent pixels. The upper left corner of the image or screen
    is (0, 0). The x-coordinate increases towards the right, and the y-coordinate
    increases towards the bottom.

    Regions are represented as tuples of (left, top, right, bottom). For example,
    a 400 by 200-pixel region in the upper left corner would be (0, 0, 400, 200).

    **Template matching**

    Template matching refers to an operation where the (potential) location of
    a smaller image is searched from a larger image. It can be used for verifying
    certain conditions or locating UI elements for desktop or web automation.

    **Requirements**

    The default installation depends on `Pillow <https://python-pillow.org/>`_
    library, which is used for general image manipulation operations.

    For more robust and faster template matching, the library can use a combination
    of `NumPy <https://numpy.org/>`_ and `OpenCV <https://opencv.org/>`_.
    They can be installed by opting in to the `cv` dependency:

    ``pip install rpaframework[cv]``

    **Examples**

    **Robot Framework**

    The `Images` library can be imported and used directly in Robot Framework,
    for instance, for capturing screenshots or verifying something on the screen.

    Desktop automation based on images should be done using the corresponding
    desktop library, i.e. ``RPA.Desktop``.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Images

        *** Keywords ***
        Should show success
            [Documentation]    Raises ImageNotFoundError if success image is not on screen
            Find template on screen    ${CURDIR}${/}success.png

        Save screenshot to results
            [Documentation]    Saves screenshot of desktop with unique name
            ${timestamp}=      Get current date    result_format=%H%M%S
            Take screenshot    filename=${OUTPUT_DIR}${/}desktop_${timestamp}.png

    **Python**

    .. code-block:: python

        from RPA.Images import Images

        def draw_matches_on_image(source, template):
            matches = lib.find_template_in_image(source, template)
            for match in matches:
                lib.show_region_in_image(source, match)

            source.save("matches.png")
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.matcher = TemplateMatcher()

    def crop_image(self, image, region, filename=None) -> None:
        """Crop an existing image.

        :param image:       Image to crop
        :param region:      Region to crop image to
        :param filename:    Save cropped image to filename
        """
        region = to_region(region)
        image = to_image(image)

        image = image.crop(region.as_tuple())
        image.load()

        if filename:
            # Suffix isn't created automatically here
            image.save(Path(filename).with_suffix(".png"), "PNG")
            notebook_image(filename)

    def find_template_in_image(
        self, image, template, region=None, limit=None, tolerance=None
    ) -> List[Region]:
        """Attempt to find the template from the given image.

        :param image:       Path to image or Image instance, used to search from
        :param template:    Path to image or Image instance, used to search with
        :param limit:       Limit returned results to maximum of `limit`.
        :param region:      Area to search from. Can speed up search significantly.
        :param tolerance:   Tolerance for matching, value between 0.1 and 1.0
        :return:            List of matching regions
        :raises ImageNotFoundError: No match was found
        :raises ValueError: Template is larger than search region
        """
        # Ensure images are in Pillow format
        image = to_image(image)
        template = to_image(template)

        # Crop image if requested
        if region is not None:
            region = to_region(region)
            image = image.crop(region.as_tuple())

        # Verify template still fits in image
        if template.size[0] > image.size[0] or template.size[1] > image.size[1]:
            raise ValueError("Template is larger than search region")

        # Strip alpha channels
        if image.mode == "RGBA":
            image = image.convert("RGB")
        if template.mode == "RGBA":
            template = template.convert("RGB")

        # Do the actual search
        start = time.time()
        matches = self.matcher.match(image, template, limit, tolerance)
        logging.info("Scanned image in %.2f seconds", time.time() - start)

        if not matches:
            raise ImageNotFoundError("No matches for given template")

        # Convert region coördinates back to full-size coördinates
        if region is not None:
            for match in matches:
                match.move(region.left, region.top)

        return matches

    def show_region_in_image(self, image, region, color="red", width=5) -> Image:
        """Draw a rectangle onto the image around the given region.

        :param image:   image to draw onto
        :param region:  coordinates for region or Region object
        :param color:   color of rectangle
        :param width:   line width of rectangle
        :return:    Image of the selected region
        """
        image = to_image(image)
        region = to_region(region)

        draw = ImageDraw.Draw(image)
        draw.rectangle(region.as_tuple(), outline=color, width=int(width))
        return image

    def get_pixel_color_in_image(self, image, point) -> RGB:
        """Get the RGB value of a pixel in the image.

        :param image:   image to get pixel from
        :param point:   coordinates for pixel or Point object
        :return: RGB value of pixel in image
        """
        point = to_point(point)
        pixel = image.getpixel(point.as_tuple())
        return RGB.from_pixel(pixel)


class TemplateMatcher:
    """Container class for different template matching methods."""

    DEFAULT_TOLERANCE = 0.95  # Tolerance for correlation matching methods
    LIMIT_FAILSAFE = 256  # Fail-safe limit of maximum match count

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self._tolerance = self.DEFAULT_TOLERANCE
        self._tolerance_warned = False

    @property
    def tolerance(self) -> float:
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value) -> float:
        self._tolerance = clamp(0.10, value, 1.00)

    def match(self, image, template, limit=None, tolerance=None) -> List[Region]:
        """Attempt to find the template in the given image.

        :param image:       image to search from
        :param template:    image to search with
        :param limit:       maximum number of returned matches
        :param tolerance:   minimum correlation factor between template and image
        :return:            list of regions that match the criteria
        """
        if HAS_RECOGNITION:
            return self._find_recognition(image, template, limit, tolerance)
        else:
            return self._find_exact(image, template, limit, tolerance)

    def _find_recognition(
        self, image, template, limit=None, tolerance=None
    ) -> List[Region]:
        """Find template using recognition module."""
        if tolerance is None:
            tolerance = self._tolerance

        confidence = tolerance * 100.0
        return templates.find(image, template, limit=limit, confidence=confidence)

    def _find_exact(self, image, template, limit=None, tolerance=None) -> List[Region]:
        """Fallback finder when no recognition module available."""
        if tolerance is not None and not self._tolerance_warned:
            self._tolerance_warned = True
            self.logger.warning(
                "Template matching tolerance not supported for current search method"
            )

        matches = []
        for match in self._iter_matches(image, template):
            matches.append(match)
            if limit is not None and len(matches) >= int(limit):
                break
            elif len(matches) >= self.LIMIT_FAILSAFE:
                self.logger.warning(
                    "Reached maximum of %d matches", self.LIMIT_FAILSAFE
                )
                break

        return matches

    def _iter_matches(self, image, template) -> Region:
        """Brute-force search for template image in larger image.

        Use optimized string search for finding the first row and then
        check if whole template matches.

        TODO: Generalize string-search algorithm to work in two dimensions
        """
        image = ImageOps.grayscale(image)
        template = ImageOps.grayscale(template)

        template_width, template_height = template.size
        template_rows = chunks(tuple(template.getdata()), template_width)

        image_width, _ = image.size
        image_rows = chunks(tuple(image.getdata()), image_width)

        for image_y, image_row in enumerate(image_rows[: -len(template_rows)]):
            for image_x in self._search_string(image_row, template_rows[0]):
                match = True
                for match_y, template_row in enumerate(template_rows[1:], image_y):
                    match_row = image_rows[match_y][image_x : image_x + template_width]
                    if template_row != match_row:
                        match = False
                        break

                if match:
                    yield Region.from_size(
                        image_x, image_y, template_width, template_height
                    )

    def _search_string(self, text, pattern) -> int:
        """Python implementation of Knuth-Morris-Pratt string search algorithm."""
        pattern_len = len(pattern)

        # Build table of shift amounts
        shifts = [1] * (pattern_len + 1)
        shift = 1
        for idx in range(pattern_len):
            while shift <= idx and pattern[idx] != pattern[idx - shift]:
                shift += shifts[idx - shift]
            shifts[idx + 1] = shift

        # Do the actual search
        start_idx = 0
        match_len = 0
        for char in text:
            while (
                match_len == pattern_len
                or match_len >= 0
                and pattern[match_len] != char
            ):
                start_idx += shifts[match_len]
                match_len -= shifts[match_len]
            match_len += 1
            if match_len == pattern_len:
                yield start_idx



from enum import Enum
import json
import logging
from pathlib import Path
from typing import Optional, Dict, List, Union, Any, Tuple

from robot.api.deco import library, keyword
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from RPA.HTTP import HTTP
from RPA.RobotLogListener import RobotLogListener

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


class ConfigurationType(Enum):
    """Possible configuration types"""

    default = "default"
    work_items = "workItemIds"
    # storages is at the moment unsupported type by the API
    # storages = "storages"


def to_configuration_type(value: Any) -> ConfigurationType:
    """Convert value to ConfigurationType enum."""
    if isinstance(value, ConfigurationType):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return ConfigurationType[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown configuration type: {value}") from err


@library
class Process:
    """A library for interacting with Control Room (CR) Process API endpoints.

    See `Operating Workforce`_ for information about process run, step run and work
    item states.

    See `APIs and webhooks`_ for information about Control Room APIs.

    .. _Operating Workforce: https://robocorp.com/docs/control-room/operating-workforce
    .. _APIs and webhooks: https://robocorp.com/docs/control-room/apis-and-webhooks

    **Examples**

    **Robot Framework**

    In the following example a task creates two input work items,
    and starts a process with those items. This results in 2 different
    process runs in the Control Room.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Robocorp.Process
        Library    RPA.Robocorp.Vault

        *** Keywords ***
        Initialize Process Library
            ${secrets}=  Get Secret  ProcessAPI
            Set Credentials
            ...   ${secrets}[workspace_id]
            ...   ${secrets}[process_id]
            ...   ${secrets}[apikey]

        *** Tasks ***
        Start process with work items
            [Setup]   Initialize Process Library
            &{item1}=  Create Dictionary  fname=Mark  lname=Monkey
            &{item2}=  Create Dictionary  fname=John  lname=Doe
            @{items}=  Create List  ${item1}   ${item2}
            Start Process  work_items=${items}  batch=True

    **Robot Framework**

    In the following example a task creates work item with files.
    To include files in a work item, the item needs to be created
    before starting the process (note. different start keyword than above).

    In this example I am using same keywords and settings from above example.

    .. code-block:: robotframework

        *** Tasks ***
        Start process with work items
            [Setup]   Initialize Process Library
            &{data}=  Create Dictionary  fname=Mark  lname=Monkey
            @{files}=  Create List
            ...   ${CURDIR}${/}workdata.xlsx
            ...   ${CURDIR}${/}other.csv
            ${item_id}=    Create Input Work Item
            ...   payload=${data}
            ...   files=${files}
            Start Configured Process
            ...  config_type=work_items
            ...  extra_info=${item_id}

    **Python**

    List work items in Control Room and retry failed items.

    .. code-block:: python

        from RPA.Robocorp.Process import Process
        from RPA.Robocorp.Vault import Vault

        secrets = Vault().get_secret("ProcessAPI")
        process = Process(
            secrets["workspace_id"],
            secrets["process_id"],
            secrets["apikey"]
        )


        def retry_failed_items():
            items = process.list_process_work_items()
            for item in items:
                if item["state"] == "FAILED":
                    print("FAILED work item: %s" % item["id"])
                    result = process.retry_work_item(item["id"])
                    print(result)

        if __name__ == "__main__":
            retry_failed_items()
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    # TODO. Add support for all Process API endpoints

    def __init__(
        self,
        workspace_id: str = None,
        process_id: str = None,
        workspace_api_key: str = None,
        **kwargs,
    ):
        self.workspace_id = None
        self.process_id = None
        self.workspace_api_key = None
        listener = RobotLogListener()
        listener.register_protected_keywords(
            [
                "RPA.Robocorp.Process.set_workspace_id",
                "RPA.Robocorp.Process.set_process_id",
                "RPA.Robocorp.Process.set_apikey",
                "RPA.Robocorp.Process.set_credentials",
            ]
        )
        self.logger = logging.getLogger(__name__)

        self.robocorp_api_server = kwargs.pop(
            "robocorp_api_server", "https://api.eu1.robocorp.com/process-v1"
        )
        self.set_credentials(workspace_id, process_id, workspace_api_key)
        self.http = HTTP()

    @keyword(tags=["set"])
    def set_workspace_id(self, workspace_id: Optional[str] = None) -> None:
        """Set Control Room workspace ID

        :param workspace_id: ID of the Control Room workspace
        """
        if workspace_id:
            self.workspace_id = workspace_id

    @keyword(tags=["set"])
    def set_process_id(self, process_id: Optional[str] = None) -> None:
        """Set Control Room process ID

        :param process_id: ID of the Control Room process
        """
        if process_id:
            self.process_id = process_id

    @keyword(tags=["set"])
    def set_apikey(self, apikey: Optional[str] = None) -> None:
        """Set Workspace API access key

        :param apikey: workspace API access key
        """
        if apikey:
            self.workspace_api_key = apikey

    @keyword(tags=["set"])
    def set_credentials(
        self,
        workspace_id: Optional[str] = None,
        process_id: Optional[str] = None,
        apikey: Optional[str] = None,
    ) -> None:
        """Set credentials needed by the Process API

        :param workspace_id: ID of the Control Room workspace
        :param process_id: ID of the Control Room process
        :param apikey: workspace API access key
        """
        self.set_workspace_id(workspace_id)
        self.set_process_id(process_id)
        self.set_apikey(apikey)

    @property
    def headers(self) -> Dict[str, str]:
        return {"Authorization": f"RC-WSKEY {self.workspace_api_key}"}

    @property
    def base_api(self) -> str:
        return f"{self.robocorp_api_server}/workspaces/{self.workspace_id}"

    def process_api(self, process_id: Optional[str] = None) -> str:
        pid = process_id or self.process_id
        return f"{self.base_api}/processes/{pid}"

    def workspace_api(self, workspace_id: Optional[str] = None) -> str:
        wid = workspace_id or self.workspace_id
        return f"{self.robocorp_api_server}/workspaces/{wid}"

    @keyword(tags=["process", "post", "work item", "start"])
    def start_process(
        self,
        work_items: Optional[Union[Dict, List[Dict]]] = None,
        batch: bool = False,
        process_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Start a Control Room process

        :param work_items: input work items for the process (default empty)
        :param batch: set to True if sending list of workitems to start each
         as a separate run
        :param process_id: specific process to start
        :return: JSON of the request response

        Table showing different results depending on parameter
        values.

        ========== ====== =======
        work_items batch  result
        ========== ====== =======
        None       False  Trigger a process with empty a work item
        None       True   Error. work_items needs to be a list
        dict       False  Trigger a process with a work item containing payload of a dict
        dict       True   Error. work_items needs to be a list
        list       False  Trigger a process with a work item containing payload of a list
        list       True   Trigger multiple process runs with work items each containing payload of a dict
        ========== ====== =======
        """  # noqa: E501
        endpoint = "runs-batch" if batch else "runs"
        response = self.http.session_less_post(
            url=f"{self.process_api(process_id)}/{endpoint}",
            headers=self.headers,
            json=work_items or [],
        )

        return response.json()

    @keyword(tags=["process", "post", "work item", "start"])
    def start_configured_process(
        self,
        config_type: ConfigurationType = ConfigurationType.default,
        extra_info: Optional[Union[str, List]] = None,
        process_id: Optional[str] = None,
    ) -> str:
        """Start a Control Room process with the provided configuration

        :param config_type: type of the start, (ConfigurationType.default)
        :param extra_info: data to be sent with the start, for example. work item IDs
        :param process_id: specific process to start
        :return: string of the request response
        """
        ctype = to_configuration_type(config_type)
        request_data = {"type": ctype.value}
        if ctype == ConfigurationType.work_items:
            extra_info = (
                [extra_info] if isinstance(extra_info, str) else extra_info or []
            )
            request_data["workItemIds"] = extra_info
        # elif ctype == ConfigurationType.storages:
        #    request_data["storages"] = extra_info
        response = self.http.session_less_post(
            url=f"{self.process_api(process_id)}/run-request",
            headers=self.headers,
            data=json.dumps(request_data),
        )
        return response.text

    @keyword(tags=["process", "post", "work item"])
    def create_input_work_item(
        self,
        payload: Optional[Any] = None,
        files: Optional[Union[str, List]] = None,
        process_id: Optional[str] = None,
    ) -> int:
        """Create an input work item for a process

        :param payload: work item data
        :param files: absolute filepaths as single string or list
        :param process_id: specific process to which item belongs to
        :return: The integer that represents the work item id
        """
        files = [files] if isinstance(files, str) else files or []
        response = self.http.session_less_post(
            url=f"{self.process_api(process_id)}/work-items",
            headers=self.headers,
            json={"payload": payload or {}},
        )
        response.raise_for_status()
        response_json = response.json()
        workitem_id = response_json["id"]
        for f in files:
            filepath = Path(f).absolute()
            filename = Path(f).name
            _, register_response = self.register_file_upload(
                filepath, filename, workitem_id
            )
            self.upload_file_to_s3(
                filepath,
                filename,
                register_response,
            )
        return workitem_id

    def register_file_upload(
        self,
        filepath: str,
        workitem_filename: str,
        workitem_id: str,
        process_id: Optional[str] = None,
    ) -> Tuple[str, Dict[str, Any]]:
        upload_filesize = Path(filepath).stat().st_size
        response = self.http.session_less_post(
            url=f"{self.process_api(process_id)}/work-items/{workitem_id}/files/upload",
            headers=self.headers,
            data=json.dumps(
                {"fileName": workitem_filename, "fileSize": upload_filesize}
            ),
        )
        response.raise_for_status()
        return response.status_code, response.json()

    def upload_file_to_s3(
        self, filepath: str, workitem_filename: str, data: Any
    ) -> Tuple[str, str]:
        with open(filepath, "rb") as infile:
            url = data["url"]
            fields = data["fields"]
            files = {"file": (workitem_filename, infile.read())}
            response = self.http.session_less_post(url, data=fields, files=files)
            response.raise_for_status()
            return response.status_code, response.text

    @keyword(tags=["process", "get"])
    def list_processes(self, workspace_id: Optional[str] = None) -> Dict[str, Any]:
        """List all processes in a workspace

        :param workspace_id: specific Control Room workspace to which process belongs to
        :return: the JSON data of the process runs based on the provided parameters
        """
        response = self.http.session_less_get(
            url=f"{self.workspace_api(workspace_id)}/processes",
            headers=self.headers,
        )
        response.raise_for_status()
        return response.json()["data"]

    @keyword(tags=["process", "get", "work item"])
    def list_process_work_items(
        self,
        process_id: Optional[str] = None,
        include_data: bool = False,
        item_state: Optional[str] = None,
    ) -> Union[str, None, Any]:
        """List work items belonging to a process

        :param include_data: include work item payload and files in
         the response (default False)
        :param item_state: state of work items to return (default all)
        :param process_id: specific process to which items belongs to
        :return: the JSON data of the process runs based on the provided parameters
        """
        response = self.http.session_less_get(
            url=f"{self.process_api(process_id)}/work-items",
            headers=self.headers,
            params={"includeData": str(include_data).lower()},
        )
        response.raise_for_status()
        data = response.json()["data"]
        return (
            [d for d in data if d["state"].upper() == item_state.upper()]
            if item_state
            else data
        )

    @keyword(tags=["process", "get", "work item"])
    def list_process_run_work_items(
        self,
        process_run_id: str = None,
        process_id: str = None,
        include_data: bool = False,
        item_state: str = None,
    ):
        """List work items belonging to a specific process run

        :param process_run_id: specific process step run to which items
         belongs to
        :param process_id: specific process to which items belongs to
        :param include_data: include work item payload and files in
         the response (default False)
        :param item_state: state of work items to return (default all)
        """
        response = self.http.session_less_get(
            url=f"{self.process_api(process_id)}/runs/{process_run_id}/work-items",
            headers=self.headers,
            params={
                "includeData": str(include_data).lower(),
                "sortBy": "createTs",
                "sortOrder": "desc",
            },
        )
        response.raise_for_status()
        data = response.json()["data"]
        return (
            [d for d in data if d["state"].upper() == item_state.upper()]
            if item_state
            else data
        )

    @keyword(tags=["process", "get", "work item"])
    def get_work_item(
        self,
        workitem_id: str,
        include_data: bool = False,
        process_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get work item from Control Room

        :param workitem_id: id of the work item
        :param include_data: include work item payload and files in
         the response (default False)
        :param process_id: specific process to which runs belongs to
        :return: the JSON of the work items associated with a given process
        """
        response = self.http.session_less_get(
            url=f"{self.process_api(process_id)}/work-items/{workitem_id}",
            headers=self.headers,
            params={"includeData": str(include_data).lower()},
        )
        response.raise_for_status()
        return response.json()

    @keyword(tags=["process", "get", "runs"])
    def list_process_runs(
        self,
        run_state: Optional[str] = None,
        limit: Optional[int] = 10,
        process_id: Optional[str] = None,
    ) -> Union[str, None, Any]:
        """List of runs related to a process

        :param run_state: state of runs to return (default all)
        :param limit: number of runs to return (default 10)
        :param process_id: specific process to which runs belongs to
        :return: the JSON data of the process runs based on the provided parameters
        """
        response = self.http.session_less_get(
            url=f"{self.process_api(process_id)}/runs",
            headers=self.headers,
            params={"limit": limit},
        )
        response.raise_for_status()
        data = response.json()["data"]
        return (
            [d for d in data if d["state"].upper() == run_state.upper()]
            if run_state
            else data
        )

    @keyword(tags=["process", "get", "runs"])
    def list_process_runs_in_workspace(
        self,
        run_state: Optional[str] = None,
        limit: Optional[int] = 10,
        workspace_id: Optional[str] = None,
    ) -> Union[str, None, Any]:
        """List all process runs in a workspace

        :param run_state: state of runs to return (default all)
        :param limit: number of runs to return (default 10)
        :param workspace_id: specific Control Room workspace to which process belongs to
        :return: the JSON data of the process runs based on the provided parameters
        """
        response = self.http.session_less_get(
            url=f"{self.workspace_api(workspace_id)}/pruns",
            headers=self.headers,
            params={"limit": limit},
        )
        response.raise_for_status()
        data = response.json()
        return (
            [d for d in data if d["state"].upper() == run_state.upper()]
            if run_state
            else data
        )

    @keyword(tags=["process", "get", "runs"])
    def get_process_run_status(
        self,
        process_run_id: str,
        step_run_id: Optional[str] = None,
        process_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get a process run status by run id

        :param process_run_id: id of the process run
        :param step_run_id: id of the process step run
        :param process_id: specific process to which runs belongs to
        :return: the response JSON
        """
        request_url = f"{self.process_api(process_id)}/runs/{process_run_id}"
        if step_run_id:
            request_url = f"{request_url}/robotRuns/{step_run_id}"
        response = self.http.session_less_get(
            url=request_url,
            headers=self.headers,
        )
        response.raise_for_status()
        return response.json()

    @keyword(tags=["process", "get"])
    def get_process_id_by_name(
        self, process_name: str, workspace_id: Optional[str] = None
    ) -> Union[int, None]:
        """Get a process id of the process by name

        :param process_name: name of the process in the Control Room
        :param workspace_id: specific Control Room workspace to which process belongs to
        :return: the next iteration id or None
        """
        processes = self.list_processes(workspace_id)
        return next((p["id"] for p in processes if p["name"] == process_name), None)

    @keyword(tags=["process", "post", "work item", "retry"])
    def retry_work_item(
        self,
        work_item_id: str,
        process_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Retry processing of work item in FAILED state

        :param work_item_id: ID of the work item to retry
        :param process_id: specific process to start
        :return: the response JSON
        """
        response = self.http.session_less_post(
            url=f"{self.process_api(process_id)}/work-items/{work_item_id}/retry",
            headers=self.headers,
        )

        return response.json()



import json
import logging
import sys
from collections import OrderedDict
from typing import Any, Union

import requests
from simple_salesforce import Salesforce as SimpleSalesforce
from simple_salesforce import SFType

from RPA.Tables import Table, Tables


class SalesforceAuthenticationError(Exception):
    "Error when authenticated Salesforce instance does not exist."


class SalesforceDataNotAnDictionary(Exception):
    "Error when parameter is not dictionary as expected."


class Salesforce:
    """`Salesforce` is a library for accessing Salesforce using REST API.
    The library extends `simple-salesforce library`_.

    More information available at `Salesforce REST API Developer Guide`_.

    .. _Salesforce REST API Developer Guide:
        https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/intro_what_is_rest_api.htm

    .. _simple-salesforce library:
        https://github.com/simple-salesforce/simple-salesforce

    **Dataloader**

    The keyword `execute_dataloader_import` can be used to mimic
    `Salesforce Dataloader`_ import behaviour.

    `input_object` can be given in different formats. Below is an example where
    input is in `RPA.Table` format in **method a** and list format in **method b**.

    .. _Salesforce Dataloader:
        https://developer.salesforce.com/docs/atlas.en-us.dataLoader.meta/dataLoader/data_loader.htm

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Salesforce
        Library     RPA.Database
        Task Setup  Authorize Salesforce

        *** Tasks ***
        # Method a
        ${orders}=        Database Query Result As Table
        ...               SELECT * FROM incoming_orders
        ${status}=        Execute Dataloader Insert
        ...               ${orders}  ${mapping_dict}  Tilaus__c
        # Method b
        ${status}=        Execute Dataloader Insert
        ...               ${WORKDIR}${/}orders.json  ${mapping_dict}  Tilaus__c


    Example file **orders.json**

    .. code-block:: json

        [
            {
                "asiakas": "0015I000002jBLIQA2"
            },
            {
                "asiakas": "0015I000002jBLDQA2"
            },
        ]

    `mapping_object` describes how the input data fields are mapped into Salesforce
    object attributes. In the example, the mapping defines that `asiakas` attribute in the
    input object is mapped into `Tilaaja__c` attribute of `Tilaus__c` custom Salesforce object.

    .. code-block:: json

        {
            "Tilaus__c": {
                "asiakas": "Tilaaja__c"
            },
        }

    Object type could be, for example, `Tilaus__c`.

    **Salesforce object operations**

    Following operations can be used to manage Salesforce objects:

    * Get Salesforce Object By Id
    * Create Salesforce Object
    * Update Salesforce Object
    * Upsert Salesforce Object
    * Delete Salesforce Object
    * Get Salesforce Object Metadata
    * Describe Salesforce Object

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Salesforce
        Task Setup  Authorize Salesforce

        *** Variables ***
        ${ACCOUNT_NOKIA}    0015I000002jBLDQA2

        *** Tasks ***
        Change account details in Salesforce
            &{account}=      Get Salesforce Object By Id   Account  ${ACCOUNT_NOKIA}
            &{update_obj}=   Create Dictionary   Name=Nokia Ltd  BillingStreet=Nokia bulevard 1
            ${result}=       Update Salesforce Object  Account  ${ACCOUNT_NOKIA}  ${update_obj}

        *** Keywords ***
        Authorize Salesforce
            ${secrets}=     Get Secret   salesforce
            Auth With Token
            ...        username=${secrets}[USERNAME]
            ...        password=${secrets}[PASSWORD]
            ...        api_token=${secrets}[API_TOKEN]

    **Python**

    .. code-block:: python

        import pprint
        from RPA.Salesforce import Salesforce
        from RPA.Robocorp.Vault import FileSecrets

        pp = pprint.PrettyPrinter(indent=4)
        filesecrets = FileSecrets("secrets.json")
        secrets = filesecrets.get_secret("salesforce")

        sf = Salesforce()
        sf.auth_with_token(
            username=secrets["USERNAME"],
            password=secrets["PASSWORD"],
            api_token=secrets["API_TOKEN"],
        )
        nokia_account_id = "0015I000002jBLDQA2"
        account = sf.get_salesforce_object_by_id("Account", nokia_account_id)
        pp.pprint(account)
        billing_information = {
            "BillingStreet": "Nokia Bulevard 1",
            "BillingCity": "Espoo",
            "BillingPostalCode": "01210",
            "BillingCountry": "Finland",
        }
        result = sf.update_salesforce_object("Account", nokia_account_id, billing_information)
        print(f"Update result: {result}")

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    account = {"Name": None, "Id": None}

    def __init__(self, sandbox: bool = False) -> None:
        self.logger = logging.getLogger(__name__)
        self.sf = None
        self.domain = "test" if sandbox else "login"
        self.session = None
        self.pricebook_name = None
        self.dataloader_success = []
        self.dataloader_errors = []

    def _require_authentication(self) -> None:
        if self.sf is None:
            raise SalesforceAuthenticationError("Authentication is not completed")

    @property
    def session_id(self):
        return self.sf.session_id if self.sf else None

    @property
    def instance(self):
        return self.sf.sf_instance if self.sf else None

    def auth_with_token(self, username: str, password: str, api_token: str) -> None:
        """Authorize to Salesforce with security token, username
        and password creating instance.

        :param username: Salesforce API username
        :param password: Salesforce API password
        :param api_token: Salesforce API security token
        """
        self.session = requests.Session()
        self.sf = SimpleSalesforce(
            username=username,
            password=password,
            security_token=api_token,
            domain=self.domain,
            session=self.session,
        )
        self.logger.debug("Salesforce session id: %s", self.session_id)

    def _get_values(self, node, prefix=None, data=None):
        if data is None:
            data = []
        if prefix is None:
            prefix = []

        if isinstance(node, OrderedDict):
            for k, v in node.items():
                if k != "attributes":
                    prefix.append(k)
                    self._get_values(v, prefix, data)
                    prefix = prefix[:-1]
        else:
            data.append((sys.intern(".".join(prefix)), node))

        return data

    def _generate_table_from_SFDC_API_query(
        self, result: dict, start: int = 0, limit: int = 0
    ) -> Table:
        records = (
            result["records"][start:]
            if limit == 0
            else result["records"][start : start + limit]
        )
        if len(records) == 0:
            return Tables().create_table()

        cols = [col for col, _ in self._get_values(records[0])]
        table = Tables().create_table(columns=cols)

        for row in records:
            values = self._get_values(row)
            table.append_row(row=dict(values))

        return table

    def salesforce_query(
        self, sql_string: str, as_table: bool = False
    ) -> Union[dict, Table]:
        """Perform SQL query and return result as `dict` or `Table`.

        :param sql_string: SQL clause to perform.
        :param as_table: Set to `True` if the result should be of `RPA.Tables.Table`
            type. (dictionary is returned by default)
        :returns: Result of the SQL query.
        """
        self._require_authentication()
        result: dict = self.sf.query_all(sql_string)
        if not as_table:
            return result

        return self._generate_table_from_SFDC_API_query(result)

    def salesforce_query_result_as_table(self, sql_string: str) -> Table:
        """Shorthand for ``Salesforce Query    ${sql_string}    as_table=${True}``.

        :param sql_string: SQL clause to perform.
        :returns: Result of the SQL query as `RPA.Tables.Table`.
        """
        return self.salesforce_query(sql_string, as_table=True)

    def set_account(self, account_name: str = "", account_id: str = "") -> bool:
        """Set account name and id by giving either parameter.

        Can be used together with keywords:
            - get_opportunity_id
            - create_new_opportunity

        :param account_name: string, defaults to ""
        :param account_id: string, defaults to ""
        :return: True if account was found from Salesforce, else False
        """
        result = self.salesforce_query(
            f"SELECT Id, Name FROM Account WHERE Name = '{account_name}' "
            f"or Id = '{account_id}'"
        )
        if result["totalSize"] == 1:
            self.account["Id"] = result["records"][0]["Id"]
            self.account["Name"] = result["records"][0]["Name"]
            self.logger.debug("Found account: %s", self.account)
            return True
        else:
            self.account = {"Name": None, "Id": None}
            return False

    def get_pricebook_entries(self) -> dict:
        """Get all pricebook entries.

        :return: query result
        """
        return self.salesforce_query("SELECT Id, Name FROM Pricebook2")

    def get_opportunity_id(self, opportunity_name: str) -> Any:
        """Get ID of an Opportunity linked to set account.

        :param opportunity_name: opportunity to query
        :return: Id of the opportunity or False
        """
        sql_query = (
            f"SELECT Id, AccountId FROM Opportunity WHERE Name = '{opportunity_name}'"
        )

        if self.account["Id"] is not None:
            sql_query += " AND AccountId = '%s'" % self.account["Id"]

        result = self.salesforce_query(sql_query)
        if result["totalSize"] == 1:
            return result["records"][0]["Id"]
        return False

    def get_pricebook_id(self, pricebook_name: str) -> Any:
        """Get ID of a pricelist.

        Returns False if unique Id is not found.

        :param pricebook_name: pricelist to query
        :return: Id of the pricelist or False
        """
        result = self.salesforce_query(
            f"SELECT Id FROM Pricebook2 WHERE Name = '{pricebook_name}'"
        )
        if result["totalSize"] == 1:
            return result["records"][0]["Id"]
        return False

    def get_products_in_pricelist(self, pricebook_name: str) -> dict:
        """Get all products in a pricelist.

        :param pricebook_name: pricelist to query
        :return: products in dictionary
        """
        result = self.salesforce_query(
            f"SELECT PriceBook2.Name, Product2.Id, Product2.Name, UnitPrice, Name "
            f"FROM PricebookEntry WHERE PriceBook2.Name = '{pricebook_name}'"
        )
        products = {}
        for item in result["records"]:
            product_name = item["Product2"]["Name"]
            pricebook_entry_id = item["attributes"]["url"].split("/")[-1]
            product_unitprice = item["UnitPrice"]
            products[product_name] = {
                "pricebook_entry_id": pricebook_entry_id,
                "unit_price": product_unitprice,
            }
        return products

    def set_pricebook(self, pricebook_name: str) -> None:
        """Sets Pricebook to be used in Salesforce operations.

        :param pricebook_name: pricelist to use
        """
        self.pricebook_name = pricebook_name

    def add_product_into_opportunity(
        self,
        product_name: str,
        quantity: int,
        opportunity_id: str = None,
        pricebook_name: str = None,
        custom_total_price: float = None,
    ) -> bool:
        """Add Salesforce Product into Opportunity.

        :param product_name: type of the product in the Pricelist
        :param quantity: number of products to add
        :param opportunity_id: identifier of Opportunity, default None
        :param pricebook_name: name of the pricelist, default None
        :param custom_total_price: price that overrides quantity and product price,
            default None
        :return: True is operation is successful or False
        """
        self._require_authentication()
        if opportunity_id is None:
            return False
        if pricebook_name:
            products = self.get_products_in_pricelist(pricebook_name)
        else:
            products = self.get_products_in_pricelist(self.pricebook_name)
        sfobject = SFType("OpportunityLineItem", self.session_id, self.instance)
        if product_name in products.keys():
            data_object = {
                "OpportunityId": opportunity_id,
                "PricebookEntryId": products[product_name]["pricebook_entry_id"],
                "Quantity": int(quantity),
                "TotalPrice": int(quantity) * products[product_name]["unit_price"],
            }
            if custom_total_price:
                data_object["TotalPrice"] = float(custom_total_price)
            result = sfobject.create(data_object)
            if result and bool(result["success"]):
                return True
        return False

    def create_new_opportunity(
        self,
        close_date: str,
        opportunity_name: str,
        stage_name: str = "Closed Won",
        account_name: str = None,
    ) -> Any:
        """Create Salesforce Opportunity object.

        :param close_date: closing date for the Opportunity, format 'YYYY-MM-DD'
        :param opportunity_name: as string
        :param stage_name: needs to be one of the defined stages,
            defaults to "Closed Won"
        :param account_name: by default uses previously set account, defaults to None
        :return: created opportunity or False
        """
        self._require_authentication()
        # "2020-04-03"
        if account_name:
            self.set_account(account_name=account_name)
        if self.account["Id"] is None:
            return False

        sfobject = SFType("Opportunity", self.session_id, self.instance)
        result = sfobject.create(
            {
                "CloseDate": close_date,
                "Name": opportunity_name,
                "StageName": stage_name,
                "Type": "Initial Subscription",
                "AccountId": self.account["Id"],
            }
        )
        self.logger.debug("create new opportunity: %s", result)
        return result.get("id") or False

    def read_dictionary_from_file(self, mapping_file: str) -> dict:
        """Read dictionary from file.

        :param mapping_file: path to the file
        :return: file content as dictionary
        """
        mapping = None
        with open(mapping_file, "r", encoding="utf-8") as mf:
            mapping = json.loads(mf.read())
        return mapping

    def _get_input_iterable(self, input_object):
        input_iterable = {}
        if isinstance(input_object, dict):
            input_iterable = input_object.items
        elif isinstance(input_object, Table):
            input_iterable = input_object.iter_dicts
        elif isinstance(input_object, list):
            input_table = Table(input_object)
            input_iterable = input_table.iter_dicts
        else:
            input_dict = self.read_dictionary_from_file(input_object)
            if isinstance(input_dict, list):
                input_table = Table(input_dict)
                input_iterable = input_table.iter_dicts
            else:
                input_iterable = input_dict
        return input_iterable

    def execute_dataloader_insert(
        self, input_object: Any, mapping_object: Any, object_type: str
    ) -> bool:
        """Keyword mimics Salesforce Dataloader 'insert' behaviour by taking
        in a `input_object`representing dictionary of data to input into Salesforce,
        a `mapping_object` representing dictionary mapping the input keys into
        Salesforce keys, an `object_type` representing Salesforce object which
        Datahandler will handle with `operation` type.

        Stores operation successes into `Salesforce.dataloader_success` array.
        Stores operation errors into `Salesforce.dataloader_errors`.

        These can be retrieved with keywords `get_dataloader_success_table` and
        `get_dataloader_error_table` which return corresponding data as
        `RPA.Table`.

        :param input_object: filepath or list of dictionaries
        :param mapping_object: filepath or dictionary
        :param object_type: Salesforce object type
        :return: True if operation is successful
        """
        self._require_authentication()
        if not isinstance(mapping_object, (dict, Table)):
            mapping_dict = self.read_dictionary_from_file(mapping_object)
        else:
            mapping_dict = mapping_object

        input_iterable = self._get_input_iterable(input_object)
        sfobject = SFType(object_type, self.session_id, self.instance)
        self.dataloader_success = []
        self.dataloader_errors = []
        for item in input_iterable():
            data_object = {}
            for key, value in mapping_dict[object_type].items():
                data_object[value] = item[key]
            result = sfobject.create(data_object)
            if result["success"]:
                data_status = {"result_id": result["id"]}
                self.dataloader_success.append({**data_status, **item})
            else:
                data_status = {"message": "failed"}
                self.dataloader_errors.append({**data_status, **item})
        return True

    def get_dataloader_success_table(self) -> Table:
        "Return Dataloader success entries as `RPA.Table`"
        return Table(self.dataloader_success)

    def get_dataloader_error_table(self) -> Table:
        "Return Dataloader error entries as `RPA.Table`"
        return Table(self.dataloader_errors)

    def get_salesforce_object_by_id(self, object_type: str, object_id: str) -> dict:
        """Get Salesforce object by id and type.

        :param object_type: Salesforce object type
        :param object_id: Salesforce object id
        :return: dictionary of object attributes
        """
        self._require_authentication()
        sfobject = SFType(object_type, self.session_id, self.instance)
        return sfobject.get(object_id)

    def create_salesforce_object(self, object_type: str, object_data: Any) -> dict:
        """Create Salesforce object by type and data.

        :param object_type: Salesforce object type
        :param object_data: Salesforce object data
        :raises SalesforceDataNotAnDictionary: when `object_data` is not dictionary
        :return: resulting object as dictionary
        """
        self._require_authentication()
        if not isinstance(object_data, dict):
            raise SalesforceDataNotAnDictionary(object_data)
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        result = salesforce_object.create(object_data)
        return dict(result)

    def update_salesforce_object(
        self, object_type: str, object_id: str, object_data: Any
    ) -> bool:
        """Update Salesfoce object by type, id and data.

        :param object_type: Salesforce object type
        :param object_id: Salesforce object id
        :param object_data: Salesforce object data
        :raises SalesforceDataNotAnDictionary: when `object_data` is not dictionary
        :return: True if successful
        """
        self._require_authentication()
        if not isinstance(object_data, dict):
            raise SalesforceDataNotAnDictionary(object_data)
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        result_code = salesforce_object.update(object_id, object_data)
        return result_code == 204

    def upsert_salesforce_object(
        self, object_type: str, object_id: str, object_data: Any
    ) -> bool:
        """Upsert Salesfoce object by type, id and data.

        :param object_type: Salesforce object type
        :param object_id: Salesforce object id
        :param object_data: Salesforce object data
        :raises SalesforceDataNotAnDictionary: when `object_data` is not dictionary
        :return: True if successful
        """
        self._require_authentication()
        if not isinstance(object_data, dict):
            raise SalesforceDataNotAnDictionary(object_data)
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        result_code = salesforce_object.upsert(object_id, object_data)
        return result_code == 204

    def delete_salesforce_object(self, object_type: str, object_id: str) -> bool:
        """Delete Salesfoce object by type and id.

        :param object_type: Salesforce object type
        :param object_id: Salesforce object id
        :return: True if successful
        """
        self._require_authentication()
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        result_code = salesforce_object.delete(object_id)
        return result_code == 204

    def get_salesforce_object_metadata(self, object_type: str) -> dict:
        """Get Salesfoce object metadata by type.

        :param object_type: Salesforce object type
        :return: object metadata as dictionary
        """
        self._require_authentication()
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        return dict(salesforce_object.metadata())

    def describe_salesforce_object(self, object_type: str) -> dict:
        """Get Salesfoce object description by type.

        :param object_type: Salesforce object type
        :return: object description as dictionary
        """
        self._require_authentication()
        salesforce_object = SFType(object_type, self.session_id, self.instance)
        return dict(salesforce_object.describe())



import datetime
import functools

import mock
import pytest
from exchangelib.services import common as exchangelib_common
from exchangelib.util import post_ratelimited

from RPA.Email.Exchange import UTC, Exchange
from RPA.Robocorp.Vault import Vault

from . import RESOURCES_DIR


SENDMAIL_MOCK = "RPA.Email.Exchange.Message.send"
recipient = "person1@domain.com"
multi_recipients = "person2@domain.com,person3@domain.com"

# FIXME: Re-test and refactor every bad unit test. (explicitly skip the ones requiring
#  online Vault and valid credentials)
pytest.skip("until all tests are fixed", allow_module_level=True)


@pytest.fixture
def library():
    lib = Exchange()
    return lib


@pytest.fixture
def local_vault(monkeypatch):
    monkeypatch.setenv("RPA_SECRET_MANAGER", "RPA.Robocorp.Vault.FileSecrets")
    monkeypatch.setenv("RPA_SECRET_FILE", str(RESOURCES_DIR / "secrets.yaml"))
    return Vault()


def _extended_post_ratelimited(protocol, session, *args, **kwargs):
    session.post = functools.partial(
        session.post,
        client_id=protocol.credentials.client_id,
        client_secret=protocol.credentials.client_secret,
    )
    return post_ratelimited(protocol, session, *args, **kwargs)


@pytest.fixture(params=[True, False])
def is_oauth(monkeypatch, request):
    if request.param:  # OAuth2 flow enabled
        # Monkey-patching infinite loop fix on token refresh, due to OAuth2 library bug
        #  when passing the app credentials to the "/token" endpoint.
        monkeypatch.setattr(
            exchangelib_common, "post_ratelimited", _extended_post_ratelimited
        )
    return request.param


@mock.patch(SENDMAIL_MOCK)
def test_send_message_no_recipients_raises(mocked, library):
    with pytest.raises(TypeError):
        library.send_message(subject="My test email subject")


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_recipients(mocked, library):
    status = library.send_message(recipients=recipient)
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_multiple_recipients(mocked, library):
    status = library.send_message(recipients=multi_recipients)
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_multiple_recipients_as_list(mocked, library):
    multi_recipients_as_list = ["person2@domain.com, person3@domain.com"]
    status = library.send_message(recipients=multi_recipients_as_list)
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_with_subject(mocked, library):
    status = library.send_message(recipients=recipient, subject="My test email subject")
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_with_body(mocked, library):
    status = library.send_message(
        recipients=recipient, body="email body of the message"
    )
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_with_html_body_but_in_plaintext(mocked, library):
    status = library.send_message(
        recipients=recipient, body="<p>email <b>body</b> of the message</p>"
    )
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_with_html_body_html_format(mocked, library):
    status = library.send_message(
        recipients=recipient, body="<p>email <b>body</b> of the message</p>", html=True
    )
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_images(mocked, library):
    status = library.send_message(
        recipients=recipient, images=RESOURCES_DIR / "approved.png"
    )
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_attachments(mocked, library):
    status = library.send_message(
        recipients=recipient, attachments=RESOURCES_DIR / "approved.png"
    )
    assert status


@mock.patch(SENDMAIL_MOCK)
def test_send_message_with_images_and_attachments(mocked, library):
    imagepath = str(RESOURCES_DIR / "approved.png")
    status = library.send_message(
        recipients=recipient, attachments=imagepath, images=imagepath
    )
    assert status


def test_get_filter_key_value_unknown_criterion_key(library):
    with pytest.raises(KeyError):
        _ = library._get_filter_key_value("dfddff")
    with pytest.raises(KeyError):
        _ = library._get_filter_key_value("dfddff:a")
    with pytest.raises(KeyError):
        _ = library._get_filter_key_value("dfddff:")


def test_get_filter_key_value_dates(library):
    default_start = datetime.datetime(1972, 1, 1, tzinfo=UTC)
    default_end = datetime.datetime(2050, 1, 1, tzinfo=UTC)
    date_1 = library._parse_date_from_string("05-05-2020")
    date_2 = library._parse_date_from_string("01-02-2021")

    criterias = {
        "before:01-02-2021": {
            "datetime_received__range": (default_start, date_2),
        },
        "after:05-05-2020": {
            "datetime_received__range": (date_1, default_end),
        },
        "between:'05-05-2020 and 01-02-2021'": {
            "datetime_received__range": (date_1, date_2),
        },
    }

    for criteria, expected in criterias.items():
        result = library._get_filter_key_value(criteria)
        assert result == expected


def test_get_filter_key_value_by_importance(library):
    result = library._get_filter_key_value("importance:high")
    assert result == {"importance": "High"}


def test_get_filter_key_value_by_category(library):
    result = library._get_filter_key_value("category:green")
    assert result == {"categories": "green"}


def test_get_filter_key_value_by_category_contains(library):
    result = library._get_filter_key_value("category_contains:green")
    assert result == {"categories__contains": "green"}


def test_get_filter_key_value_by_body(library):
    result = library._get_filter_key_value("body:'Content in the email body'")
    assert result == {"body": "Content in the email body"}


def test_get_filter_key_value_by_body_contains(library):
    result = library._get_filter_key_value("body_contains:'email body'")
    assert result == {"body__contains": "email body"}


def test_get_filter_key_value_by_subject(library):
    result = library._get_filter_key_value("subject:'Hello World'")
    assert result == {"subject": "Hello World"}


def test_get_filter_key_value_by_subject_contains(library):
    result = library._get_filter_key_value("subject_contains:buy")
    assert result == {"subject__contains": "buy"}


def test_get_filter_key_value_by_sender(library):
    result = library._get_filter_key_value("sender:name@domain.com")
    assert result == {"sender": "name@domain.com"}


def test_get_filter_key_value_by_sender_contains(library):
    result = library._get_filter_key_value("sender_contains:name@domain.com")
    assert result == {"sender__contains": "name@domain.com"}


def test_get_filter_by_key_value_multiple_conditions(library):
    criterias = {
        "sender:robocorp.tester@gmail.com and subject:epic": {
            "sender": "robocorp.tester@gmail.com",
            "subject": "epic",
        },
        "sender_contains:robocorp.tester@gmail.com and subject:'epic level'": {
            "sender__contains": "robocorp.tester@gmail.com",
            "subject": "epic level",
        },
        # first quoted criteria is prioritized
        "body_contains:message and body_contains:'epic level'": {
            "body__contains": "epic level",
        },
        # first quoted criteria is prioritized
        "body_contains:'message' and body_contains:'epic level'": {
            "body__contains": "message",
        },
    }

    for criteria, expected in criterias.items():
        result = library._get_filter_key_value(criteria)
        assert result == expected


@pytest.mark.skip(reason="requires online Vault and valid e-mail account")
@pytest.mark.parametrize("to_self", [True, False])
def test_send_message_with_out_bcc_addresses(library, local_vault, is_oauth, to_self):
    secrets = local_vault.get_secret("exchange")
    if is_oauth:
        params = {
            "access_type": "IMPERSONATE",
            "client_id": secrets["client_id"],
            "client_secret": secrets["client_secret"],
            "token": secrets["token"],
        }
    else:
        params = {
            "password": secrets["password"],
        }
    library.authorize(
        username=secrets["account"],
        autodiscover=False,
        server="outlook.office365.com",
        is_oauth=is_oauth,
        **params,
    )
    if to_self:
        recipient = secrets["account"]
        bcc_list = []
        body = "Sent to self without BCC"
    else:
        recipient = None
        bcc_list = ["robocorp.tester@gmail.com", "robocorp.tester.2@gmail.com"]
        body = "Sent to others with BCC only"
    library.send_message(
        recipients=recipient,
        bcc=bcc_list,
        subject="test_send_message_with_out_bcc_addresses",
        body=f"{body} ({'OAuth2' if is_oauth else 'basic'})",
    )



import importlib
import logging
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from RPA.RobotLogListener import RobotLogListener
from RPA.Tables import Table

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


class Configuration:
    """Class to handle configuration from config files
    and class init"""

    def __init__(self):
        self.configuration = {}
        self.module_name = None

    def parse_arguments(
        self,
        module_name,
        database,
        username,
        password,
        host,
        port,
        charset,
        config_file: str,
    ):
        config = ConfigParser.ConfigParser()
        config.read([config_file])

        self.configuration = {}
        self.module_name = module_name or (
            config.get("default", "module_name")
            if config.has_option("default", "module_name")
            else None
        )
        self.configuration["database"] = database or (
            config.get("default", "database")
            if config.has_option("default", "database")
            else None
        )
        self.configuration["username"] = username or (
            config.get("default", "username")
            if config.has_option("default", "username")
            else None
        )
        self.configuration["password"] = password or (
            config.get("default", "password")
            if config.has_option("default", "password")
            else None
        )
        self.configuration["host"] = host or (
            config.get("default", "host")
            if config.has_option("default", "host")
            else None
        )
        self.configuration["port"] = port or (
            int(config.get("default", "port"))
            if config.has_option("default", "port")
            else None
        )
        self.configuration["charset"] = charset or (
            config.get("default", "charset")
            if config.has_option("default", "charset")
            else None
        )
        return self.module_name, self.configuration

    def get(self, param, default=None):
        # Missing values are still present in configuration as nulls.
        value = self.configuration.get(param)
        return value if value is not None else default

    def set_val(self, param, value):
        self.configuration[param] = value

    def all_but_empty(self):
        new_dict = {}
        for key, value in dict(self.configuration).items():
            if value is not None:
                new_dict[key] = value
        return new_dict

    def set_default_port(self, port):
        if (
            "port" not in self.configuration.keys()
            or self.configuration["port"] is None
        ):
            self.configuration["port"] = int(port)

    def get_connection_parameters_as_string(self, conf=None):
        configuration = conf or self.configuration
        parameters = ",".join(
            "{}={}".format(str(k), str(v)) for k, v in configuration.items()
        )
        return "Connecting using : %s.connect(%s)" % (self.module_name, parameters)


class Database:
    """`Database` is a library for handling different database operations.

    All database operations are supported. Keywords `Query` and `Get Rows`
    return values by default in `RPA.Table` format.

    Library is compatible with any Database API Specification 2.0 module.

    References:

    - Database API Specification 2.0 - http://www.python.org/dev/peps/pep-0249/
    - Lists of DB API 2.0 - http://wiki.python.org/moin/DatabaseInterfaces
    - Python Database Programming - http://wiki.python.org/moin/DatabaseProgramming/

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library         RPA.Database

        *** Tasks ***
        Get Orders From Database
            Connect To Database  pymysql  tester  user  password  127.0.0.1
            @{orders}            Query    Select * FROM incoming_orders
            FOR   ${order}  IN  @{orders}
                Handle Order  ${order}
            END

    **Python**

    .. code-block:: python

        from RPA.Database import Database
        from RPA.Robocorp.Vault import FileSecrets

        filesecrets = FileSecrets("secrets.json")
        secrets = filesecrets.get_secret("databasesecrets")

        db = Database()
        db.connect_to_database('pymysql',
                            secrets["DATABASE"],
                            secrets["USERNAME"],
                            secrets["PASSWORD"],
                            '127.0.0.1'
                            )
        orders = db.query("SELECT * FROM incoming_orders")
        for order in orders:
            print(order)
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self._dbconnection = None
        self.db_api_module_name = None
        self.config = Configuration()
        listener = RobotLogListener()
        listener.register_protected_keywords(["RPA.Database.connect_to_database"])

    # pylint: disable=R0915, too-many-branches
    def connect_to_database(  # noqa: C901
        self,
        module_name: Optional[str] = None,
        database: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        charset: Optional[str] = None,
        config_file: Optional[str] = "db.cfg",
        autocommit: Optional[bool] = False,
    ) -> None:
        """Connect to database using DB API 2.0 module.

        :param module_name: database module to use
        :param database: name of the database
        :param username: of the user accessing the database
        :param password: of the user accessing the database
        :param host: SQL server address
        :param port: SQL server port
        :param charset: for example, "utf-8", defaults to None
        :param config_file: location of configuration file, defaults to "db.cfg"
        :param autocommit: set autocommit value for connect (only with pymssql atm)

        Example:

        .. code-block:: robotframework

            Connect To Database  pymysql  database  username  password  host  port
            Connect To Database  ${CURDIR}${/}resources${/}dbconfig.cfg

        """
        # TODO. take autocommit into use for all database modules
        self.config.parse_arguments(
            module_name, database, username, password, host, port, charset, config_file
        )
        if self.config.module_name in ("excel", "excelrw"):
            self.db_api_module_name = "pyodbc"
            dbmodule = importlib.import_module("pyodbc")
        else:
            self.db_api_module_name = self.config.module_name
            dbmodule = importlib.import_module(self.config.module_name)
        if module_name in ["MySQLdb", "pymysql"]:
            self.config.set_default_port(3306)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                db=self.config.get("database"),
                user=self.config.get("username"),
                passwd=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
                charset=self.config.get("charset"),
            )
        elif module_name == "psycopg2":
            self.config.set_default_port(5432)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                database=self.config.get("database"),
                user=self.config.get("username"),
                password=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
            )
        elif module_name in ("pyodbc", "pypyodbc"):
            self.config.set_default_port(1433)
            server = self.config.get("host", "")
            if server:
                server += f",{self.config.get('port')}"
            db = self.config.get("database", "")
            usr = self.config.get("username", "")
            pwd = self.config.get("password", "")
            self.config.set_val(
                "connect_string",
                f"DRIVER={{SQL Server}};SERVER={server};DATABASE={db};"
                f"UID={usr};PWD={pwd};",
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(self.config.get("connect_string"))
        elif module_name == "excel":
            self.config.set_val(
                "connect_string",
                "DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};"
                'DBQ=%s;ReadOnly=1;Extended Properties="Excel 8.0;HDR=YES";)'
                % self.config.get("database"),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                autocommit=True,
            )
        elif module_name == "excelrw":
            self.config.set_val(
                "connect_string",
                "DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};"
                'DBQ=%s;ReadOnly=0;Extended Properties="Excel 8.0;HDR=YES";)'
                % self.config.get("database"),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                autocommit=True,
            )
        elif module_name in ("ibm_db", "ibm_db_dbi"):
            self.config.set_default_port(50000)
            self.config.set_val(
                "connect_string",
                "DATABASE=%s;HOSTNAME=%s;PORT=%s;PROTOCOL=TCPIP;UID=%s;PWD=%s;"
                % (
                    self.config.get("database"),
                    self.config.get("host"),
                    self.config.get("port"),
                    self.config.get("username"),
                    self.config.get("password"),
                ),
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                self.config.get("connect_string"),
                "",
                "",
            )
        elif module_name in ("cx_Oracle", "oracledb"):
            self.config.set_default_port(1521)
            oracle_dsn = dbmodule.makedsn(
                host=self.config.get("host"),
                port=self.config.get("port"),
                service_name=self.config.get("database"),
            )
            self.config.set_val("oracle_dsn", oracle_dsn)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                user=self.config.get("username"),
                password=self.config.get("password"),
                dsn=self.config.get("oracle_dsn"),
            )
        elif module_name == "teradata":
            self.config.set_default_port(1025)
            teradata_udaExec = dbmodule.UdaExec(
                appName="RobotFramework", version="1.0", logConsole=False
            )
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = teradata_udaExec.connect(
                method="odbc",
                system=self.config.get("host"),
                database=self.config.get("database"),
                username=self.config.get("username"),
                password=self.config.get("password"),
                host=self.config.get("host"),
                port=self.config.get("port"),
            )
        elif module_name == "pymssql":
            self.config.set_default_port(1433)
            self.logger.info(self.config.get_connection_parameters_as_string())
            self._dbconnection = dbmodule.connect(
                server=self.config.get("host"),
                user=self.config.get("username"),
                password=self.config.get("password"),
                database=self.config.get("database"),
                port=self.config.get("port"),
                host=self.config.get("host", "."),
                autocommit=autocommit,
            )
        else:
            conf = self.config.all_but_empty()
            self.logger.info(self.config.get_connection_parameters_as_string(conf))
            self._dbconnection = dbmodule.connect(**conf)
            if module_name == "sqlite3":
                self._dbconnection.isolation_level = None if autocommit else "IMMEDIATE"

    def call_stored_procedure(
        self,
        name: str,
        params: Optional[List[str]] = None,
        sanstran: Optional[bool] = False,
    ) -> List[str]:
        """Call stored procedure with name and params.

        :param name: procedure name
        :param params: parameters for the procedure as a list, defaults to None
        :param sanstran: Run the query without an implicit transaction commit or
            rollback if such additional action was detected. (turned off by default)
        :returns: list of results

        Example:

        .. code-block:: robotframework

            @{params}     Create List   FirstParam   SecondParam   ThirdParam
            @{results}    Call Stored Procedure   mystpr  ${params}

        """
        params = params or []
        cur = None
        try:
            if self.db_api_module_name in ("cx_Oracle", "oracledb"):
                cur = self._dbconnection.cursor()
            else:
                cur = self._dbconnection.cursor(as_dict=False)
            PY3K = sys.version_info >= (3, 0)
            if not PY3K:
                name = name.encode("ascii", "ignore")
            cur.callproc(name, params)
            cur.nextset()
            value = []
            for row in cur:
                value.append(row)
        except Exception as exc:
            # Implicitly rollback when error occurs.
            self.logger.error(exc)
            if cur and not sanstran:
                self._dbconnection.rollback()
            raise
        else:
            if not sanstran:
                self._dbconnection.commit()
            return value

    def description(self, table: str) -> list:
        """Get description of the SQL table

        :param table: name of the SQL table
        :returns: database descripton as a list

        Example:

        .. code-block:: robotframework

            Connect To Database    pymysql  mydb  user  pass  127.0.0.1
            ${db_description}      Description  mytable

        """
        try:
            result = self.query("DESCRIBE %s" % table, as_table=True)
        except Exception as e:
            raise AssertionError(
                "Operation not supported for '%s' type database"
                % self.db_api_module_name
            ) from e
        return result.to_list()

    def disconnect_from_database(self) -> None:
        """Close connection to SQL database

        Example:

        .. code-block:: robotframework

            Connect To Database    pymysql  mydb  user  pass  127.0.0.1
            ${result}              Query   Select firstname, lastname FROM table
            Disconnect From Database

        """
        if self._dbconnection:
            self._dbconnection.close()

    # pylint: disable=R0912
    def execute_sql_script(  # noqa: C901
        self,
        filename: str,
        sanstran: Optional[bool] = False,
        encoding: Optional[str] = "utf-8",
    ) -> None:  # noqa: C901
        """Execute content of SQL script as SQL commands.

        :param filename: filepath to SQL script to execute
        :param sanstran: Run the query without an implicit transaction commit or
            rollback if such additional action was detected. (turned off by default)
        :param encoding: character encoding of file (utf-8 by default)

        Example:

        .. code-block:: robotframework

            Execute SQL Script   script.sql

        """
        with open(filename, encoding=encoding) as script_file:
            sql_script = script_file.readlines()

        cur = None
        try:
            cur = self._dbconnection.cursor()
            sqlStatement = ""
            for line in sql_script:
                if line.startswith("#") or line.startswith("--"):
                    continue

                sql_fragments = line.split(";")
                if len(sql_fragments) == 1:
                    sqlStatement += line + " "
                else:
                    for sqlFragment in sql_fragments:
                        sqlFragment = sqlFragment.strip()
                        if len(sqlFragment) == 0:
                            continue

                        sqlStatement += sqlFragment + " "

                        self.__execute_sql(cur, sqlStatement)
                        sqlStatement = ""

            sqlStatement = sqlStatement.strip()
            if len(sqlStatement) != 0:
                self.__execute_sql(cur, sqlStatement)
        except Exception as exc:
            # Implicitly rollback when error occurs.
            self.logger.error(exc)
            if cur and not sanstran:
                self._dbconnection.rollback()
            raise
        else:
            if not sanstran:
                self._dbconnection.commit()

    def query(
        self,
        statement: str,
        assertion: Optional[str] = None,
        sanstran: Optional[bool] = False,
        as_table: Optional[bool] = True,
        returning: Optional[bool] = None,
    ) -> Union[List, Dict, Table, Any]:
        """Execute a SQL query and optionally return the execution result.

        :param statement: SQL statement to execute.
        :param assertion: Assert on query result, row_count or columns.
            Works only for `SELECT` statements. (defaults to `None`)
        :param sanstran: Run the query without an implicit transaction commit or
            rollback if such additional action was detected and this is set to `True`.
            (turned off by default, meaning that *commit* is performed on successful
            queries and *rollback* on failing ones automatically)
        :param as_table: If the result should be an instance of `Table`, otherwise a
            `list` will be returned. (defaults to `True`)
        :param returning: Set this to `True` if you want to have rows explicitly
            returned (instead of the query result), `False` otherwise. (by default a
            heuristic detects if it should return or not)
        :returns: Fetched rows when `returning` is `True` or if the heuristic decides
            that the statement should return (raw rows or as `Table` if `as_table` is
            `True`), otherwise the object produced by the execution is returned.

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            *** Settings ***
            Library    RPA.Database

            *** Tasks ***
            Select Values From Table
                @{rows} =    Query   SELECT id,value FROM table
                FOR  ${row}  IN  @{rows}
                    Log   ${row}
                END
                @{res} =    Query   Select * FROM table   row_count > ${EXPECTED}
                @{res} =    Query   Select * FROM table   'value' in columns
                @{res} =    Query   Select * FROM table   columns == ['id', 'value']

        **Python**

        .. code-block:: python

            from RPA.Database import Database

            lib = Database()

            def insert_and_return_names():
                lib.connect_to_database("sqlite3", "sqlite.db")
                lib.query("DROP TABLE IF EXISTS orders;")
                lib.query("CREATE TABLE orders(id INTEGER PRIMARY KEY, name TEXT);")
                rows = lib.query(
                    'INSERT INTO orders(id, name) VALUES(1, "my-1st-order"),'
                    '(2, "my-2nd-order") RETURNING name;'
                )
                print([row["name"] for row in rows])  # ['my-1st-order', 'my-2nd-order']
        """
        cursor = None
        try:
            self.logger.info("Executing query: %s", statement)
            cursor = self._dbconnection.cursor()
            result = self.__execute_sql(cursor, statement)
            should_return = (returning is True) or (
                returning is None and self._is_returnable_statement(statement)
            )
            if should_return:
                rows = [tuple(row) for row in cursor.fetchall()]
                columns = [col[0] for col in (cursor.description or [])]
                self._result_assertion(rows, columns, assertion)
                if as_table:
                    result = Table(rows, columns)
                else:
                    result = rows
        except Exception as exc:
            # Implicitly rollback when error occurs.
            self.logger.error(exc)
            if cursor and not sanstran:
                self._dbconnection.rollback()
            raise
        else:
            if not sanstran:
                self._dbconnection.commit()
            return result

    def _is_returnable_statement(self, statement: str) -> bool:
        lower_parts = statement.lower().split()

        starts_with = lower_parts[0]
        if starts_with in ["select", "describe", "show", "explain"]:
            return True

        if "returning" in lower_parts:
            return True

        return False

    @staticmethod
    def _result_assertion(rows: List[Tuple[Any]], columns: List[str], assertion: str):
        if not assertion:
            return

        # pylint: disable=unused-variable
        row_count = len(rows)  # noqa: F841
        available_locals = {
            "row_count": row_count,
            "columns": columns,
        }
        # pylint: disable=W0123
        valid = eval(assertion, {"__builtins__": None}, available_locals)

        if not valid:
            raise AssertionError(
                "Query assertion %s failed. Facts: %s" % (assertion, available_locals)
            )

    def __execute_sql(self, cursor, sqlStatement):
        return cursor.execute(sqlStatement)

    def set_auto_commit(self, autocommit: bool = True) -> None:
        """Set database auto commit mode.

        :param autocommit: boolean value for auto commit, defaults to True

        Example:

        .. code-block:: robotframework

            Set Auto Commit             # auto commit is set on
            Set Auto Commit   False     # auto commit is turned off

        """
        self._dbconnection.autocommit = autocommit

    def get_rows(
        self,
        table,
        columns: Optional[str] = None,
        conditions: Optional[str] = None,
        as_table: Optional[bool] = True,
    ) -> Union[List, Dict, Table, Any]:
        """Get rows from table. Columns and conditions can be
        set to filter result.

        :param table: name of the SQL table
        :param columns: name of columns to return, defaults to `None`
         means that all columns are returned
        :param conditions: limiting result by WHERE clause, defaults to `None`
        :param as_table: if result should be instance of ``Table``, defaults to `True`
         `False` means that return type would be `list`
        :returns: table or list based on param as_table arguement

        Example:

        .. code-block:: robotframework

            @{res}   Get Rows  tablename  arvo
            @{res}   Get Rows  tablename  arvo  columns=id,name
            @{res}   Get Rows  tablename  columns=id  conditions=column1='newvalue'
            @{res}   Get Rows  tablename  conditions=column2='updatedvalue'

        """
        columns = columns or "*"
        where_cond = f" WHERE {conditions}" if conditions else ""
        return self.query(
            "SELECT %s FROM %s%s" % (columns, table, where_cond), as_table=as_table
        )

    def get_number_of_rows(self, table: str, conditions: Optional[str] = None) -> int:
        """Get number of rows in a table. Conditions can be given
        as arguments for WHERE clause.

        :param table: name of the SQL table
        :param conditions: restrictions for selections, defaults to None
        :returns: number or rows

        Example:

        .. code-block:: robotframework

            ${count}   Get Number Of Rows  tablename
            ${count}   Get Number Of Rows  tablename  column1=5 and column2='x'

        """
        where_cond = f" WHERE {conditions}" if conditions else ""
        result = self.query(
            "SELECT COUNT(*) FROM %s%s" % (table, where_cond), as_table=False
        )
        return result[0][0]



from datetime import datetime, timedelta
import json
import sys
import click

from robot.libraries.OperatingSystem import OperatingSystem
from RPA.FileSystem import FileSystem

PROD_API_URL = "https://api.eu1.robocloud.eu"


def run_command(command):
    ret = OperatingSystem().run(command)
    return ret


def output_and_exit(message, level="error"):
    bg_color = "red"
    fg_color = "white"
    msg_prefix = "Error"
    if level == "warn":
        bg_color = "yellow"
        fg_color = "black"
        msg_prefix = "Warning"
    click.secho(f"{msg_prefix}:", bg=bg_color, fg=fg_color, nl=False)
    click.echo(f" {message}")
    sys.exit(0)


def does_dev_access_account_exist():
    creds = run_command("rcc configure credentials --json")
    creds = json.loads(creds)
    return "DEV" in [ac["account"] for ac in creds]


def check_for_environment_variables():
    variables = OperatingSystem().get_environment_variables()
    env_vars = {
        "RC_API_SECRET_HOST": variables.get("RC_API_SECRET_HOST", None),
        "RC_WORKSPACE_ID": variables.get("RC_WORKSPACE_ID", None),
        "RC_API_SECRET_TOKEN": variables.get("RC_API_SECRET_TOKEN", None),
    }
    return env_vars


def ask_for_access_credentials():
    credentials_url = "https://cloud.robocorp.com/settings/access-credentials"
    open_url = click.confirm(
        f"Open {credentials_url} to add new credential?", default=True
    )
    if open_url:
        click.launch(credentials_url)
    access_credentials = click.prompt(
        "Create access credentials in https://cloud.robocorp.com "
        "and paste that here to create DEV account",
    )
    ret = run_command(f"rcc configure credentials {access_credentials} -a DEV")
    if "Error: " in ret:
        click.echo(ret)
    else:
        click.echo("OK")


def update_env_json(workspace, token):
    response = click.confirm(
        "Do you want to update environment variables in devdata/env.json ?",
        default=True,
    )
    if response:
        content = OperatingSystem().get_file("devdata/env.json")
        if len(content) == 0:
            content = {}
        else:
            try:
                content = json.loads(content)
            except json.decoder.JSONDecodeError:
                output_and_exit(
                    "Can't parse env.json contents. Check the file structure."
                )

        content["RC_WORKSPACE_ID"] = workspace
        content["RC_API_SECRET_TOKEN"] = token
        content["RC_API_SECRET_HOST"] = PROD_API_URL
        with open("devdata/env.json", "w", encoding="utf-8") as fout:
            fout.write(json.dumps(content, sort_keys=False, indent=4))


def create_env_json(workspace, token):
    response = click.confirm(
        "devdata/env.json does not exist - do you want to create it ?", default=True
    )
    if response:
        OperatingSystem().create_file("devdata/env.json")
        content = {}
        content["RC_WORKSPACE_ID"] = workspace
        content["RC_API_SECRET_TOKEN"] = token
        content["RC_API_SECRET_HOST"] = PROD_API_URL
        with open("devdata/env.json", "w", encoding="utf-8") as fout:
            fout.write(json.dumps(content, sort_keys=False, indent=4))


@click.group()
def cli() -> None:
    """
    Set development environment to use Robocorp Vault
    """


@cli.command(name="set")
def set_command():
    """Access Robocorp Vault in development enviroment"""
    click.secho("Use Robocorp Vault", fg="white", bold=True, underline=True)
    dev_account_exist = does_dev_access_account_exist()
    if dev_account_exist:
        click.echo("DEV access account already exists")
    else:
        ask_for_access_credentials()
    env_vars = check_for_environment_variables()
    workspace = click.prompt(
        "What is your workspace ID?",
        default=env_vars["RC_WORKSPACE_ID"],
        show_default=True,
    )

    # token_validity = click.prompt(
    #     "Access token validity time in minutes (max 1440 min = 24 hours)",
    #     default=1440,
    #     type=int,
    # )
    token_validity = 1440
    ret = run_command(f"rcc cloud authorize -a DEV -w {workspace} -m {token_validity}")
    if "Error:" in ret:
        output_and_exit(ret)
    ret = json.loads(ret)
    expiry_time = ret["when"] + timedelta(minutes=token_validity).total_seconds()

    json_exists = FileSystem().does_file_exist("devdata/env.json")
    if json_exists:
        update_env_json(workspace, ret["token"])
    else:
        create_env_json(workspace, ret["token"])

    click.echo(f"Token expires at {datetime.fromtimestamp(expiry_time)}")


@cli.command(name="clear")
def clear_command():
    """Clear DEV access account"""
    dev_account_exist = does_dev_access_account_exist()
    if dev_account_exist:
        response = click.confirm("Do you want to delete DEV account", default=True)
        if response:
            run_command("rcc configure credentials --delete -a DEV")
            click.echo("DEV account deleted")
    else:
        click.echo("DEV access account does not exist.")


@cli.command(name="list")
def list_command():
    """List workspaces"""
    dev_account_exist = does_dev_access_account_exist()
    if dev_account_exist:
        response = run_command("rcc cloud workspace -a DEV")
        response = json.loads(response)
        click.secho("Workspace name / ID", fg="white", bold=True, underline=True)
        for r in response:
            click.secho(f"{r['name']}", fg="bright_blue", nl=False)
            click.echo(" / ", nl=False)
            click.secho(f"{r['id']}", fg="green")
    else:
        output_and_exit(
            "DEV access account does not exist. Cannot list workspaces.", level="warn"
        )


def main():
    cli()


if __name__ == "__main__":
    main()
